#!usr/bin/env python3
# -*- coding:utf-8 -*-

__author__ = 'mayanqiong'

from tqsdk.tradeable.mixin import StockMixin
from tqsdk.datetime import _timestamp_nano_to_str
from tqsdk.report import TqReport
from tqsdk.tradeable.sim.basesim import BaseSim
from tqsdk.tradeable.sim.trade_stock import SimTradeStock


class TqSimStock(BaseSim, StockMixin):
    """
    天勤股票模拟交易类

    该类实现了一个本地的股票模拟交易账户，并且在内部完成撮合交易，在回测模式下，只能使用 TqSimStock 账户来交易股票合约。

    股票模拟交易只支持 ins_class 字段为 'STOCK' 的合约，且不支持 T+0 交易。

    限价单要求报单价格达到或超过对手盘价格才能成交, 成交价为报单价格, 如果没有对手盘(涨跌停)则无法成交

    市价单使用对手盘价格成交, 如果没有对手盘(涨跌停)则自动撤单

    模拟交易不会有部分成交的情况, 要成交就是全部成交

    TqSimStock 暂不支持设置手续费
    """

    def __init__(self, init_balance: float = 10000000.0, account_id: str = None) -> None:
        """
        Args:
            init_balance (float): [可选]初始资金, 默认为一千万

            account_id (str): [可选]帐号, 默认为 TQSIM_STOCK

        Example1::

            # 修改TqSim模拟帐号的初始资金为100000
            from tqsdk import TqApi, TqSimStock, TqAuth
            api = TqApi(TqSimStock(init_balance=100000), auth=TqAuth("快期账户", "账户密码"))

        Example2::

            # 同时使用 TqSim 交易期货，TqSimStock 交易股票
            from tqsdk import TqApi, TqAuth, TqMultiAccount, TqSim, TqSimStock

            tqsim_future = TqSim()
            tqsim_stock = TqSimStock()

            api = TqApi(account=TqMultiAccount([tqsim_future, tqsim_stock]), auth=TqAuth("快期账户", "账户密码"))

            # 多账户下单，需要指定下单账户
            order1 = api.insert_order(symbol="SHFE.cu2112", direction="BUY", offset="OPEN", volume=10, limit_price=72250.0, account=tqsim_future)
            order2 = api.insert_order(symbol="SSE.603666", direction="BUY", volume=300, account=tqsim_stock)
            while order1.status != 'FINISHED' or order2.status != 'FINISHED':
                api.wait_update()

            # 打印账户可用资金
            future_account = tqsim_future.get_account()
            stock_account = tqsim_stock.get_account()
            print(future_account.available, stock_account.available)
            api.close()

        Example3::

            # 在回测模式下，同时使用 TqSim 交易期货，TqSimStock 交易股票

            tqsim_future = TqSim()
            tqsim_stock = TqSimStock()
            api = TqApi(account=TqMultiAccount([tqsim_future, tqsim_stock]),
                        backtest=TqBacktest(start_dt=datetime(2021, 7, 12), end_dt=datetime(2021, 7, 14)),
                        auth=TqAuth("快期账户", "账户密码"))

            future_quote = api.get_quote("SHFE.cu2112")
            future_stock = api.get_quote("SSE.603666")

            while datetime.strptime(future_stock.datetime, "%Y-%m-%d %H:%M:%S.%f") < datetime(2021, 7, 12, 9, 50):
                api.wait_update()

            # 开仓，多账户下单，需要指定下单账户
            order1 = api.insert_order(symbol="SHFE.cu2112", direction="BUY", offset="OPEN", volume=10, limit_price=future_quote.ask_price1, account=tqsim_future)
            order2 = api.insert_order(symbol="SSE.603666", direction="BUY", volume=300, account=tqsim_stock)
            while order1.status != 'FINISHED' or order2.status != 'FINISHED':
                api.wait_update()

            future_account = tqsim_future.get_account()
            stock_account = tqsim_stock.get_account()
            # 打印账户当前可用资金
            print(future_account.available, stock_account.available)

            # 等待行情回测到第二天
            while datetime.strptime(future_stock.datetime, "%Y-%m-%d %H:%M:%S.%f") < datetime(2021, 7, 13, 10, 30):
                api.wait_update()
            # 平仓，股票只能 T+1 交易
            order3 = api.insert_order(symbol="SHFE.cu2112", direction="SELL", offset="CLOSE", volume=8, limit_price=future_quote.bid_price1, account=tqsim_future)
            order4 = api.insert_order(symbol="SSE.603666", direction="SELL", volume=200, account=tqsim_stock)
            while order3.status != 'FINISHED' or order4.status != 'FINISHED':
                api.wait_update()

            try:  # 等到回测结束
                while True:
                    api.wait_update()
            except BacktestFinished:
                # 打印回测时间内账户交易信息统计结果
                print(tqsim_future.tqsdk_stat)
                print(tqsim_stock.tqsdk_stat)
                api.close()

        """
        if float(init_balance) <= 0:
            raise Exception("初始资金(init_balance) %s 错误, 请检查 init_balance 是否填写正确" % (init_balance))
        super(TqSimStock, self).__init__(account_id="TQSIM_STOCK" if account_id is None else account_id,
                                         init_balance=float(init_balance),
                                         trade_class=SimTradeStock)

    @property
    def _account_info(self):
        info = super(TqSimStock, self)._account_info
        info.update({
            "account_type": self._account_type
        })
        return info

    @property
    def _account_auth(self):
        return {
            "feature": "sec",
            "account_id": None,
            "auto_add": False,
        }

    def _handle_on_alive(self, msg, order):
        """
        在 order 状态变为 ALIVE 调用，屏幕输出信息，打印日志
        """
        symbol = f"{order['exchange_id']}.{order['instrument_id']}"
        self._api._print(
            f"模拟交易下单 {self._account_name}, {order['order_id']}: 时间: {_timestamp_nano_to_str(order['insert_date_time'])}, "
            f"合约: {symbol}, 方向: {order['direction']}, 手数: {order['volume_left']}, "
            f"价格: {order.get('limit_price', '市价')}")
        self._logger.debug(msg, order_id=order["order_id"], datetime=order["insert_date_time"],
                           symbol=symbol, direction=order["direction"],
                           volume_left=order["volume_left"], limit_price=order.get("limit_price", "市价"))

    def _handle_on_finished(self, msg, order):
        """
        在 order 状态变为 FINISHED 调用，屏幕输出信息，打印日志
        """
        self._api._print(f"模拟交易委托单 {self._account_name}, {order['order_id']}: {order['last_msg']}")
        self._logger.debug(msg, order_id=order["order_id"], last_msg=order["last_msg"], status=order["status"],
                           volume_orign=order["volume_orign"], volume_left=order["volume_left"])

    def _report(self):
        if not self.trade_log:
            return
        date_keys = sorted(self.trade_log.keys())
        self._api._print(f"模拟交易成交记录, 账户: {self._account_name}")
        for d in date_keys:
            for t in self.trade_log[d]["trades"]:
                symbol = t["exchange_id"] + "." + t["instrument_id"]
                self._api._print(f"时间: {_timestamp_nano_to_str(t['trade_date_time'])}, 合约: {symbol}, "
                                 f"方向: {t['direction']}, 手数: {t['volume']}, 价格: {t['price']:.3f}, 手续费: {t['fee']:.2f}")

        self._api._print(f"模拟交易账户资金, 账户: {self._account_name}")
        for d in date_keys:
            account = self.trade_log[d]["account"]
            self._api._print(
                f"日期: {d}, 账户资产: {account['asset']:.2f}, 分红: {account['dividend_balance_today']:.2f}, "
                f"买入成本: {account['cost']:.2f}, 盈亏: {account['profit_today']:.2f}, 盈亏比: {account['profit_rate_today']:.2f}, "
                f"手续费: {account['buy_fee_today'] + account['sell_fee_today']:.2f}")
        report = TqReport(report_id=self._account_id, trade_log=self.trade_log, quotes=self._data['quotes'], account_type="SPOT")
        self.tqsdk_stat = report.default_metrics
        self._api._print(
            f"收益率: {self.tqsdk_stat['ror'] * 100:.2f}%, 年化收益率: {self.tqsdk_stat['annual_yield'] * 100:.2f}%, "
            f"最大回撤: {self.tqsdk_stat['max_drawdown'] * 100:.2f}%, 年化夏普率: {self.tqsdk_stat['sharpe_ratio']:.4f},"
            f"年化索提诺比率: {self.tqsdk_stat['sortino_ratio']:.4f}")



#!usr/bin/env python3
# -*- coding:utf-8 -*-
__author__ = 'yanqiong'


from tqsdk.api import TqApi
from tqsdk.diff import _get_obj


class TqNotify(object):
    """
    用于在同步代码中接受服务器通知
    """

    def __init__(self, api: TqApi) -> None:
        """
        创建 TqNotify 实例

        Args:
            api (tqsdk.api.TqApi): TqApi 实例

        Example::

            from tqsdk import TqApi, TqAuth, TqKq, TqNotify

            api = TqApi(account=TqKq(), auth=TqAuth("快期账户", "账户密码"))
            tqNotify = TqNotify(api)  # 构造实例类
            while True:
                api.wait_update()
                # 每次调用返回距离上一次调用 tqNotify.get_notifies() 之后产生的通知列表，没有的话返回 []
                notify_list = tqNotify.get_notifies()
                for notify in notify_list:
                    print(notify)  # 打印出通知内容
                    # send_message(notify['content'])  可以发送通知到其他工具

        """
        self._api = api
        self._notify = _get_obj(self._api._data, ["notify"])
        # 用户未读取过的通知，用 list 类型尽量保证用户读到通知的顺序和进程收到的顺序一致，但是不能完全保证
        self._unread_notifies_list = [k for k in self._notify if not k.startswith("_")]
        # 已经添加到 _unread_notifies 的通知
        self._processed_notifies_set = {k for k in self._notify if not k.startswith("_")}
        self._task = self._api.create_task(self._run())

    async def _run(self):
        async with self._api.register_update_notify(self._notify) as update_chan:
            async for _ in update_chan:
                all_notifies = {k for k in self._notify if not k.startswith("_")}
                notifies = all_notifies - self._processed_notifies_set  # 最近更新的通知
                self._processed_notifies_set = all_notifies
                self._unread_notifies_list.extend([k for k in notifies])

    def get_notifies(self):
        notifies = [self._notify[n] for n in self._unread_notifies_list]
        self._unread_notifies_list = []
        return notifies



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'hongyan'

from tqsdk import TqApi, TqAuth, TqAccount, TqKq, TqSim, TqMultiAccount
# 多账户模式下, 同时操作实盘、模拟交易和快期模拟账户交易
tqact = TqAccount("H海通期货", "123456", "123456")
sim = TqSim()
kq = TqKq()

with TqApi(TqMultiAccount([tqact, sim, kq]), auth=TqAuth("快期账户", "账户密码")) as api:
    order1 = api.insert_order(symbol="DCE.m2101", direction="BUY", offset="OPEN", volume=5, account=tqact)
    order2 = api.insert_order(symbol="SHFE.au2012C308", direction="BUY", offset="OPEN", volume=5, limit_price=78.1, account=sim)
    order3 = api.insert_order(symbol="SHFE.cu2101", direction="Sell", offset="OPEN", volume=10, limit_price=51610, account=kq)
    api.cancel_order(order3, kq)
    while order1.status != "FINISHED" or order2.status != "FINISHED":
        api.wait_update()
    # 分别获取账户资金信息
    account_info1 = tqact.get_account()
    account_info2 = sim.get_account()
    account_info3 = kq.get_account()
    # 分别获取账户持仓信息
    position1 = tqact.get_position("DCE.m2101", )
    position2 = sim.get_position()
    position3 = kq.get_position()
    # 分别获取账户委托数据
    orders1 = tqact.get_order(order_id=order1.order_id, )
    orders2 = sim.get_position()
    orders3 = kq.get_position()
    # 分别获取账户成交数据
    trades1 = tqact.get_trade()
    trades2 = sim.get_trade()
    trades3 = kq.get_trade()


#!/usr/bin/env python
#  -*- coding: utf-8 -*-
"""
包含有关时间处理的功能函数
"""

__author__ = 'limin'

import datetime
import time
from typing import Union

# 北京时间时区 CST (China Standard Time)
_cst_tz = datetime.timezone(datetime.timedelta(hours=8))


def _cst_now():
    """返回当前北京时间 datetime.datetime 类型"""
    return datetime.datetime.now(tz=_cst_tz)


def _convert_to_cst_datetime(user_dt: Union[datetime.date, datetime.datetime]):
    # 将用户输入的时间转换为东八区时间
    if isinstance(user_dt, datetime.datetime):
        if user_dt.tzinfo == _cst_tz:
            return user_dt
        return user_dt.replace(tzinfo=_cst_tz) if user_dt.tzinfo is None else user_dt.astimezone(tz=_cst_tz)
    else:
        return datetime.datetime(user_dt.year, user_dt.month, user_dt.day, tzinfo=_cst_tz)


def _convert_user_input_to_nano(start_dt: Union[datetime.date, datetime.datetime],
                                end_dt: Union[datetime.date, datetime.datetime]) -> (int, int):
    # 将用户输入的时间转换为对应交易时间段的纳秒时间戳
    if isinstance(start_dt, datetime.datetime):
        start_nano = _datetime_to_timestamp_nano(_convert_to_cst_datetime(start_dt))
    elif isinstance(start_dt, datetime.date):
        start_nano = _get_trading_day_start_time(_datetime_to_timestamp_nano(_convert_to_cst_datetime(start_dt)))
    else:
        raise Exception(f"start_dt 参数类型 {type(start_dt)} 错误, 只支持 datetime / date 类型，请检查是否正确")
    if isinstance(end_dt, datetime.datetime):
        end_nano = _datetime_to_timestamp_nano(_convert_to_cst_datetime(end_dt))
    elif isinstance(end_dt, datetime.date):
        end_nano = _get_trading_day_end_time(_datetime_to_timestamp_nano(_convert_to_cst_datetime(end_dt)))
    else:
        raise Exception(f"end_dt 参数类型 {type(end_dt)} 错误, 只支持 datetime / date 类型，请检查是否正确")
    return start_nano, end_nano


def _datetime_to_timestamp_nano(dt: datetime.datetime) -> int:
    # tqsdk 内部时间必须为东八区时间
    if dt.tzinfo != _cst_tz:
        dt = dt.replace(tzinfo=_cst_tz) if dt.tzinfo is None else dt.astimezone(tz=_cst_tz)
    # timestamp() 返回值精度为 microsecond，直接乘以 1e9 可能有精度问题
    return int(dt.timestamp() * 1000000) * 1000


def _timestamp_nano_to_datetime(nano: int) -> datetime.datetime:
    return datetime.datetime.fromtimestamp((nano // 1000) / 1000000, tz=_cst_tz)


def _timestamp_nano_to_str(nano: int, fmt="%Y-%m-%d %H:%M:%S.%f") -> str:
    return datetime.datetime.fromtimestamp((nano // 1000) / 1000000, tz=_cst_tz).strftime(fmt)


def _str_to_timestamp_nano(current_datetime: str, fmt="%Y-%m-%d %H:%M:%S.%f") -> int:
    return _datetime_to_timestamp_nano(datetime.datetime.strptime(current_datetime, fmt))


def _get_trading_day_start_time(trading_day):
    """给定交易日, 获得交易日起始时间"""
    begin_mark = 631123200000000000  # 1990-01-01
    start_time = trading_day - 21600000000000  # 6小时
    week_day = (start_time - begin_mark) // 86400000000000 % 7
    if week_day >= 5:
        start_time -= 86400000000000 * (week_day - 4)
    return start_time


def _get_trading_day_end_time(trading_day):
    """给定交易日, 获得交易日结束时间"""
    return trading_day + 64799999999999  # 18小时


def _get_trading_day_from_timestamp(timestamp):
    """给定时间, 获得其所属的交易日"""
    begin_mark = 631123200000000000  # 1990-01-01
    days = (timestamp - begin_mark) // 86400000000000  # 自 1990-01-01 以来的天数
    if (timestamp - begin_mark) % 86400000000000 >= 64800000000000:  # 大于18点, 天数+1
        days += 1
    week_day = days % 7
    if week_day >= 5:  # 如果是周末则移到星期一
        days += 7 - week_day
    return begin_mark + days * 86400000000000


def _get_trading_timestamp(quote, current_datetime: str):
    """ 将 quote 在 current_datetime 所在交易日的所有可交易时间段转换为纳秒时间戳(tqsdk内部使用的时间戳统一为纳秒)并返回 """
    # 获取当前交易日时间戳
    current_trading_day_timestamp = _get_trading_day_from_timestamp(_str_to_timestamp_nano(current_datetime))
    # 获取上一交易日时间戳
    last_trading_day_timestamp = _get_trading_day_from_timestamp(
        _get_trading_day_start_time(current_trading_day_timestamp) - 1)
    trading_timestamp = {
        "day": _get_period_timestamp(current_trading_day_timestamp, quote["trading_time"].get("day", [])),
        "night": _get_period_timestamp(last_trading_day_timestamp, quote["trading_time"].get("night", []))
    }
    return trading_timestamp


def _get_period_timestamp(real_date_timestamp, period_str):
    """
    real_date_timestamp：period_str 所在真实日期的纳秒时间戳（如 period_str 为周一(周二)的夜盘,则real_date_timestamp为上周五(周一)的日期; period_str 为周一的白盘,则real_date_timestamp为周一的日期）
    period_str: quote["trading_time"]["day"] or quote["trading_time"]["night"]
    """
    period_timestamp = []
    for duration in period_str:  # 对于白盘（或夜盘）中的每一个可交易时间段
        start = [int(i) for i in duration[0].split(":")]  # 交易时间段起始点
        end = [int(i) for i in duration[1].split(":")]  # 交易时间段结束点
        period_timestamp.append([real_date_timestamp + (start[0] * 3600 + start[1] * 60 + start[2]) * 1000000000,
                                 real_date_timestamp + (end[0] * 3600 + end[1] * 60 + end[2]) * 1000000000])
    return period_timestamp


def _is_in_trading_time(quote, current_datetime, local_time_record):
    """ 判断是否在可交易时间段内，需在quote已收到行情后调用本函数"""
    # 只在需要用到可交易时间段时(即本函数中)才调用_get_trading_timestamp()
    trading_timestamp = _get_trading_timestamp(quote, current_datetime)
    now_ns_timestamp = _get_trade_timestamp(current_datetime, local_time_record)  # 当前预估交易所纳秒时间戳
    # 判断当前交易所时间（估计值）是否在交易时间段内
    for v in trading_timestamp.values():
        for period in v:
            if period[0] <= now_ns_timestamp < period[1]:
                return True
    return False


def _get_trade_timestamp(current_datetime, local_time_record):
    # 根据最新行情时间获取模拟的(预估的)当前交易所纳秒时间戳（tqsdk内部使用的时间戳统一为纳秒）
    # 如果local_time_record为nan，則不加时间差
    cur_nano = _str_to_timestamp_nano(current_datetime)
    mock_delay_nano = 0 if local_time_record != local_time_record else int((time.time() - local_time_record) * 1e6) * 1000
    return cur_nano + mock_delay_nano


def _get_expire_rest_days(expire_dt, current_dt):
    """
    获取当前时间到下市时间之间的天数
    expire_dt, current_dt 都以 s 为单位
    """
    delta = datetime.datetime.fromtimestamp(expire_dt, tz=_cst_tz).date() - datetime.datetime.fromtimestamp(current_dt, tz=_cst_tz).date()
    return delta.days



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
import math
from tqsdk.entity import Entity
from tqsdk.diff import _simple_merge_diff, _get_obj


class TqStockProfit():
    """
    股票盈亏计算模块

    * 订阅已有持仓股票合约和行情
    * 计算股票持仓与资产的盈亏

    """
    def __init__(self, api):
        self._api = api
        self._data = Entity()  # 业务信息截面
        self._data._instance_entity([])
        self._diffs = []
        self._all_subscribe = set()


    async def _run(self, api_send_chan, api_recv_chan, md_send_chan, md_recv_chan):

        self._logger = self._api._logger.getChild("TqStockProfit")
        self._api_send_chan = api_send_chan
        self._api_recv_chan = api_recv_chan
        self._md_send_chan = md_send_chan
        self._md_recv_chan = md_recv_chan
        md_task = self._api.create_task(self._md_handler())
        self._pending_peek = False
        try:
            async for pack in api_send_chan:
                if "_md_recv" in pack:
                    await self._md_recv(pack)
                    await self._send_diff()
                    if not self._is_diff_complete():
                        await self._md_send_chan.send({"aid": "peek_message"})
                elif pack["aid"] == "subscribe_quote":
                    await self._subscribe_quote(set(pack["ins_list"].split(",")))
                elif pack["aid"] == "peek_message":
                    self._pending_peek = True
                    await self._send_diff()
                    if self._pending_peek:
                        await self._md_send_chan.send(pack)
                else:
                    await self._md_send_chan.send(pack)
        finally:
            await self._api._cancel_task(md_task)


    async def _md_handler(self):
        """0 接收上游数据包 """
        async for pack in self._md_recv_chan:
            pack["_md_recv"] = True
            await self._api_send_chan.send(pack)


    async def _md_recv(self, pack):
        """ 处理下行数据包
        0 将行情数据和交易数据合并至 self._data
        1 生成增量业务截面, 该截面包含 持仓盈亏和资产盈亏信息
        """
        for d in pack.get("data", {}):
            if "quotes" in d:
                # 行情数据仅仅合并沪深两市的行情数据
                stock_quote = {k: v for k, v in d.get('quotes').items() if k.startswith("SSE") or k.startswith("SZSE")}
                _simple_merge_diff(self._data, {"quotes": stock_quote})
            if "trade" in d:
                _simple_merge_diff(self._data, d)
            # 添加至 self._diff 等待被发送
            self._diffs.append(d)

        # 计算持仓和账户资产的盈亏增量截面
        pend_diff = await self._generate_pend_diff()
        self._diffs.append(pend_diff)


    async def _generate_pend_diff(self):
        """" 盈亏计算 """
        pend_diff = {}
        pend_diff.setdefault('trade', {k: {'accounts': {'CNY': {}}, 'positions': {}} for k in self._data.get('trade', {})})
        # 计算持仓盈亏
        for account_key in self._data.get('trade', {}):
            # 盈亏计算仅仅计算股票账户
            if self._data['trade'].get(account_key, {}).get("account_type", "FUTURE") == "FUTURE":
                continue
            for symbol, _ in self._data['trade'][account_key].get('positions', {}).items():
                await self._subscribe_quote(symbol)
                last_price = self._data["quotes"].get(symbol, {}).get('last_price', float("nan"))
                if not math.isnan(last_price):
                    diff = self._update_position(account_key, symbol, last_price)
                    pend_diff['trade'][account_key]['positions'][symbol] = diff
                    _simple_merge_diff(self._data["trade"][account_key]["positions"], {symbol: diff})

        # 当截面完整时, 全量刷新所有账户的资产盈亏
        if self._is_diff_complete():
            for account_key in self._data.get('trade', {}):
                if self._data['trade'].get(account_key, {}).get("account_type", "FUTURE") == "FUTURE":
                    continue
                all_position =self._data["trade"][account_key].get("positions", {})
                pend_diff['trade'][account_key]['accounts']['CNY']['float_profit'] = \
                    sum([v.get('float_profit', 0) for k, v in all_position.items()])

        return pend_diff

    async def _send_diff(self):
        if self._pending_peek and self._is_diff_complete() and self._diffs:
            rtn_data = {
                "aid": "rtn_data",
                "data": self._diffs,
            }
            self._diffs = []
            self._pending_peek = False
            await self._api_recv_chan.send(rtn_data)

    async def _subscribe_quote(self, symbols: [set, str]):
        """这里只会增加订阅合约，不会退订合约"""
        symbols = symbols if isinstance(symbols, set) else {symbols}
        if symbols - self._all_subscribe:
            self._all_subscribe |= symbols
            await self._md_send_chan.send({
                "aid": "subscribe_quote",
                "ins_list": ",".join(self._all_subscribe)
            })


    def _update_position(self, key, symbol, last_price):
        """更新持仓盈亏"""
        diff = {}
        position = self._data["trade"][key]["positions"][symbol]
        diff["last_price"] = last_price
        diff["cost"] = position['cost_price'] * position['volume']
        diff["float_profit"] = (last_price - position['cost_price']) * position['volume']
        return diff


    def _is_diff_complete(self):
        """当前账户截面是否已经完全处理完整, 即当所有股票的最新价不为空时"""
        for account_key in self._data.get('trade', {}):
            for symbol, _ in self._data['trade'][account_key].get('positions', {}).items():
                quote = self._data["quotes"].get(symbol, {})
                if math.isnan(quote.get('last_price', float("nan"))):
                    return False
        return True



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'ringo'

from tqsdk import TqApi, TqAuth
from tqsdk.ta import OPTION_IMPV

api = TqApi(auth=TqAuth("快期账户", "账户密码"))

# 获取指定期权行情
quote = api.get_quote("SHFE.cu2006C50000")

# 获取期权和对应标的的多合约 kline
klines = api.get_kline_serial(["SHFE.cu2006C50000", "SHFE.cu2006"], 24 * 60 * 60, 20)

# 通过 OPTION_IMPV 函数计算隐含波动率，设置无风险利率为 0.025
impv = OPTION_IMPV(klines, quote, 0.025)

print(list(impv["impv"] * 100))

api.close()



#!usr/bin/env python3
# -*- coding:utf-8 -*-

__author__ = 'mayanqiong'


import asyncio
import math
from datetime import date, datetime
from typing import Union, Any, List, Dict

from tqsdk.backtest.utils import TqBacktestContinuous, TqBacktestDividend
from tqsdk.channel import TqChan
from tqsdk.datetime import _get_trading_day_start_time, _get_trading_day_end_time, _get_trading_day_from_timestamp, \
    _timestamp_nano_to_str, _convert_user_input_to_nano
from tqsdk.diff import _merge_diff, _get_obj
from tqsdk.entity import Entity
from tqsdk.exceptions import BacktestFinished
from tqsdk.objs import Kline, Tick
from tqsdk.rangeset import _rangeset_range_union, _rangeset_difference, _rangeset_union
from tqsdk.utils import _generate_uuid, _query_for_quote


class BtQuote(Entity):
    """ Quote 是一个行情对象 """
    def __init__(self, api):
        self._api = api
        self.price_tick: float = float("nan")


class TqBacktest(object):
    """
    天勤回测类

    将该类传入 TqApi 的构造函数, 则策略就会进入回测模式。

    回测模式下 k线会在刚创建出来时和结束时分别更新一次, 在这之间 k线是不会更新的。

    回测模式下 quote 的更新频率由所订阅的 tick 和 k线周期确定:
        * 只要订阅了 tick, 则对应合约的 quote 就会使用 tick 生成, 更新频率也和 tick 一致, 但 **只有下字段** :
          datetime/ask&bid_price1/ask&bid_volume1/last_price/highest/lowest/average/volume/amount/open_interest/
          price_tick/price_decs/volume_multiple/max&min_limit&market_order_volume/underlying_symbol/strike_price

        * 如果没有订阅 tick, 但是订阅了 k线, 则对应合约的 quote 会使用 k线生成, 更新频率和 k线的周期一致， 如果订阅了某个合约的多个周期的 k线,
          则任一个周期的 k线有更新时, quote 都会更新. 使用 k线生成的 quote 的盘口由收盘价分别加/减一个最小变动单位, 并且 highest/lowest/average/amount
          始终为 nan, volume 始终为0

        * 如果即没有订阅 tick, 也没有订阅k线或 订阅的k线周期大于分钟线, 则 TqBacktest 会 **自动订阅分钟线** 来生成 quote

        * 如果没有订阅 tick, 但是订阅了 k线, 则对应合约的 quote **只有下字段** :
          datetime/ask&bid_price1/ask&bid_volume1/last_price/open_interest/
          price_tick/price_decs/volume_multiple/max&min_limit&market_order_volume/underlying_symbol/strike_price

    **注意** ：如果未订阅 quote，模拟交易在下单时会自动为此合约订阅 quote ，根据回测时 quote 的更新规则，如果此合约没有订阅K线或K线周期大于分钟线 **则会自动订阅一个分钟线** 。

    模拟交易要求报单价格大于等于对手盘价格才会成交, 例如下买单, 要求价格大于等于卖一价才会成交, 如果不能立即成交则会等到下次行情更新再重新判断。

    回测模式下 wait_update 每次最多推进一个行情时间。

    回测结束后会抛出 BacktestFinished 例外。

    对 **组合合约** 进行回测时需注意：只能通过订阅 tick 数据来回测，不能订阅K线，因为K线是由最新价合成的，而交易所发回的组合合约数据中无最新价。
    """

    def __init__(self, start_dt: Union[date, datetime], end_dt: Union[date, datetime]) -> None:
        """
        创建天勤回测类，起始时间和结束时间都应该是北京时间

        Args:
            start_dt (date/datetime): 回测起始时间
                * date: 指的是交易日

                * datetime: 指的是具体时间点，如果没有指定时区信息，则默认为北京时间

            end_dt (date/datetime): 回测结束时间
                * date: 指的是交易日

                * datetime: 指的是具体时间点，如果没有指定时区信息，则默认为北京时间
        """
        self._start_dt, self._end_dt = _convert_user_input_to_nano(start_dt, end_dt)
        self._current_dt = self._start_dt
        # 记录当前的交易日 开始时间/结束时间
        self._trading_day = _get_trading_day_from_timestamp(self._current_dt)
        self._trading_day_start = _get_trading_day_start_time(self._trading_day)
        self._trading_day_end = _get_trading_day_end_time(self._trading_day)

    async def _run(self, api, sim_send_chan, sim_recv_chan, md_send_chan, md_recv_chan):
        """回测task"""
        self._api = api
        # 下载历史主连合约信息
        start_trading_day = _get_trading_day_from_timestamp(self._start_dt)  # 回测开始交易日
        end_trading_day = _get_trading_day_from_timestamp(self._end_dt)  # 回测结束交易日
        self._continuous_table = TqBacktestContinuous(start_dt=start_trading_day,
                                                      end_dt=end_trading_day,
                                                      headers=self._api._base_headers)
        self._stock_dividend = TqBacktestDividend(start_dt=start_trading_day,
                                                  end_dt=end_trading_day,
                                                  headers=self._api._base_headers)
        self._logger = api._logger.getChild("TqBacktest")  # 调试信息输出
        self._sim_send_chan = sim_send_chan
        self._sim_recv_chan = sim_recv_chan
        self._md_send_chan = md_send_chan
        self._md_recv_chan = md_recv_chan
        self._pending_peek = False
        self._data = Entity()  # 数据存储
        self._data._instance_entity([])
        self._prototype = {
            "quotes": {
                "#": BtQuote(self._api),  # 行情的数据原型
            },
            "klines": {
                "*": {
                    "*": {
                        "data": {
                            "@": Kline(self._api),  # K线的数据原型
                        }
                    }
                }
            },
            "ticks": {
                "*": {
                    "data": {
                        "@": Tick(self._api),  # Tick的数据原型
                    }
                }
            }
        }
        self._sended_to_api = {}  # 已经发给 api 的 rangeset  (symbol, dur)，只记录了 kline
        self._serials = {}  # 所有用户请求的 chart 序列，如果用户订阅行情，默认请求 1 分钟 Kline
        # key 为 (ins, dur)
        # value 为 dict，包含以下属性
        #   timestamp: 当前 serial 最新的时间戳
        #   diff: 当前 serial 生成的 diff,
        #   kline_or_tick: 当前 serial 计算行情对应的 kline 或者 tick
        #   when: 'TICK' 表示由 tick 产生的行情，'OPEN' | 'CLOSE' 是 kline 开盘或者收盘时的行情
        #   chart_id_set: 记录当前 serial 对应的所有 chart_id

        # gc 是会循环 self._serials，来计算用户需要的数据，self._serials 不应该被删除，
        self._generators = {}  # 所有用户请求的 chart 序列相应的 generator 对象，创建时与 self._serials 一一对应，会在一个序列计算到最后一根 kline 时被删除
        self._had_any_generator = False  # 回测过程中是否有过 generator 对象
        self._sim_recv_chan_send_count = 0  # 统计向下游发送的 diff 的次数，每 1w 次执行一次 gc
        self._quotes = {}  # 记录 min_duration 记录某一合约的最小duration； sended_init_quote 是否已经过这个合约的初始行情
        self._diffs: List[Dict[str, Any]] = []
        self._is_first_send = True
        md_task = self._api.create_task(self._md_handler())
        try:
            await self._send_snapshot()
            async for pack in self._sim_send_chan:
                if pack["aid"] == "ins_query":
                    await self._md_send_chan.send(pack)
                    # 回测 query 不为空时需要ensure_query
                    # 1. 在api初始化时会发送初始化请求（2.5.0版本开始已经不再发送初始化请求），接着会发送peek_message，如果这里没有等到结果，那么在收到 peek_message 的时候，会发现没有数据需要发送，回测结束
                    # 2. api在发送请求后，会调用 wait_update 更新数据，如果这里没有等到结果，行情可能会被推进
                    # query 为空时，表示清空数据的请求，这个可以直接发出去，不需要等到收到回复
                    if pack["query"] != "":
                        await self._ensure_query(pack)
                    await self._send_diff()
                elif pack["aid"] == "subscribe_quote":
                    # todo: 回测时，用户如果先订阅日线，再订阅行情，会直接返回以日线 datetime 标识的行情信息，而不是当前真正的行情时间
                    self._diffs.append({
                        "ins_list": pack["ins_list"]
                    })
                    for ins in pack["ins_list"].split(","):
                        await self._ensure_quote(ins)
                    await self._send_diff()  # 处理上一次未处理的 peek_message
                elif pack["aid"] == "set_chart":
                    if pack["ins_list"]:
                        # 回测模块中已保证每次将一个行情时间的数据全部发送给api，因此更新行情时 保持与初始化时一样的charts信息（即不作修改）
                        self._diffs.append({
                            "charts": {
                                pack["chart_id"]: {
                                    # 回测中 ready 置为 True，保证api在回测中直接通过对数据接收完全的验证
                                    "left_id": -1,
                                    "right_id": -1,
                                    "ready": True,
                                    "more_data": False,  # 直接发送False给api，表明数据发送完全，使api中通过数据接收完全的验证
                                    "state": pack
                                }
                            }
                        })
                        await self._ensure_serial(pack["ins_list"], pack["duration"], pack["chart_id"])
                    else:
                        self._diffs.append({
                            "charts": {
                                pack["chart_id"]: None
                            }
                        })
                    await self._send_diff()  # 处理上一次未处理的 peek_message
                elif pack["aid"] == "peek_message":
                    self._pending_peek = True
                    await self._send_diff()
        finally:
            # 关闭所有 generator
            for s in self._generators.values():
                await s.aclose()
            await self._api._cancel_task(md_task)

    async def _md_handler(self):
        async for pack in self._md_recv_chan:
            await self._md_send_chan.send({
                "aid": "peek_message"
            })
            recv_quotes = False
            for d in pack.get("data", []):
                _merge_diff(self._data, d, self._prototype, persist=False, reduce_diff=False)
                # 收到的 quotes 转发给下游
                quotes = d.get("quotes", {})
                if quotes:
                    recv_quotes = True
                    quotes = self._update_valid_quotes(quotes)  # 删去回测 quotes 不应该下发的字段
                    self._diffs.append({"quotes": quotes})
                # 收到的 symbols 应该转发给下游
                if d.get("symbols"):
                    self._diffs.append({"symbols": d["symbols"]})
            # 如果没有收到 quotes（合约信息），或者当前的 self._data.get('quotes', {}) 里没有股票，那么不应该向 _diffs 里添加元素
            if recv_quotes:
                quotes_stock = self._stock_dividend._get_dividend(self._data.get('quotes', {}), self._trading_day)
                if quotes_stock:
                    self._diffs.append({"quotes": quotes_stock})

    def _update_valid_quotes(self, quotes):
        # 从 quotes 返回只剩余合约信息的字段的 quotes，防止发生未来数据发送给下游
        # backtest 模块会生成的数据
        invalid_keys = {f"{d}{i+1}" for d in ['ask_price', 'ask_volume', 'bid_price', 'bid_volume'] for i in range(5)}
        invalid_keys.union({'datetime', 'last_price', 'highest', 'lowest', 'average', 'volume', 'amount', 'open_interest'})
        invalid_keys.union({'cash_dividend_ratio', 'stock_dividend_ratio'})  # 这两个字段完全由 self._stock_dividend 负责处理
        # backtest 模块不会生成的数据，下游服务也不应该收到的数据
        invalid_keys.union({'open', 'close', 'settlement', 'lowest', 'lower_limit', 'upper_limit', 'pre_open_interest', 'pre_settlement', 'pre_close', 'expired'})
        for symbol, quote in quotes.items():
            [quote.pop(k, None) for k in invalid_keys]
            if symbol.startswith("KQ.m"):
                quote.pop("underlying_symbol", None)
            if quote.get('expire_datetime'):
                # 先删除所有的 quote 的 expired 字段，只在有 expire_datetime 字段时才会添加 expired 字段
                quote['expired'] = quote.get('expire_datetime') * 1000000000 <= self._trading_day_start
        return quotes

    async def _send_snapshot(self):
        """发送初始合约信息"""
        async with TqChan(self._api, last_only=True) as update_chan:  # 等待与行情服务器连接成功
            self._data["_listener"].add(update_chan)
            while self._data.get("mdhis_more_data", True):
                await update_chan.recv()
        # 发送初始行情(合约信息截面)时
        quotes = {}
        for ins, quote in self._data["quotes"].items():
            if not ins.startswith("_"):
                trading_time = quote.get("trading_time", {})
                quotes[ins] = {
                    "open": None,  # 填写None: 删除api中的这个字段
                    "close": None,
                    "settlement": None,
                    "lower_limit": None,
                    "upper_limit": None,
                    "pre_open_interest": None,
                    "pre_settlement": None,
                    "pre_close": None,
                    "ins_class": quote.get("ins_class", ""),
                    "instrument_id": quote.get("instrument_id", ""),
                    "exchange_id": quote.get("exchange_id", ""),
                    "margin": quote.get("margin"),  # 用于内部实现模拟交易, 不作为api对外可用数据（即 Quote 类中无此字段）
                    "commission": quote.get("commission"),  # 用于内部实现模拟交易, 不作为api对外可用数据（即 Quote 类中无此字段）
                    "price_tick": quote["price_tick"],
                    "price_decs": quote["price_decs"],
                    "volume_multiple": quote["volume_multiple"],
                    "max_limit_order_volume": quote["max_limit_order_volume"],
                    "max_market_order_volume": quote["max_market_order_volume"],
                    "min_limit_order_volume": quote["min_limit_order_volume"],
                    "min_market_order_volume": quote["min_market_order_volume"],
                    "underlying_symbol": quote["underlying_symbol"],
                    "strike_price": quote["strike_price"],
                    "expired": quote.get('expire_datetime', float('nan')) * 1000000000 <= self._trading_day_start,  # expired 默认值就是 False
                    "trading_time": {"day": trading_time.get("day", []), "night": trading_time.get("night", [])},
                    "expire_datetime": quote.get("expire_datetime"),
                    "delivery_month": quote.get("delivery_month"),
                    "delivery_year": quote.get("delivery_year"),
                    "option_class": quote.get("option_class", ""),
                    "product_id": quote.get("product_id", ""),
                }
        # 修改历史主连合约信息
        cont_quotes = self._continuous_table._get_history_cont_quotes(self._trading_day)
        for k, v in cont_quotes.items():
            quotes.setdefault(k, {})  # 实际上，初始行情截面中只有下市合约，没有主连
            quotes[k].update(v)
        self._diffs.append({
            "quotes": quotes,
            "ins_list": "",
            "mdhis_more_data": False,
            "_tqsdk_backtest": self._get_backtest_time()
        })

    async def _send_diff(self):
        """发送数据到 api, 如果 self._diffs 不为空则发送 self._diffs, 不推进行情时间, 否则将时间推进一格, 并发送对应的行情"""
        if self._pending_peek:
            await self._generator_diffs(True if self._diffs else False)
            if self._diffs:
                # 发送数据集中添加 backtest 字段，开始时间、结束时间、当前时间，表示当前行情推进是由 backtest 推进
                self._diffs.append({"_tqsdk_backtest": self._get_backtest_time()})

                # 切换交易日，将历史的主连合约信息添加的 diffs
                if self._current_dt > self._trading_day_end:
                    # 使用交易日结束时间，每个交易日切换只需要计算一次交易日结束时间
                    # 相比发送 diffs 前每次都用 _current_dt 计算当前交易日，计算次数更少
                    self._trading_day = _get_trading_day_from_timestamp(self._current_dt)
                    self._trading_day_start = _get_trading_day_start_time(self._trading_day)
                    self._trading_day_end = _get_trading_day_end_time(self._trading_day)
                    self._diffs.append({
                        "quotes": self._continuous_table._get_history_cont_quotes(self._trading_day)
                    })
                    self._diffs.append({
                        "quotes": self._stock_dividend._get_dividend(self._data.get('quotes'), self._trading_day)
                    })
                    self._diffs.append({
                        "quotes": {k: {'expired': v.get('expire_datetime', float('nan')) * 1000000000 <= self._trading_day_start}
                                   for k, v in self._data.get('quotes').items()}
                    })

                self._sim_recv_chan_send_count += 1
                if self._sim_recv_chan_send_count > 10000:
                    self._sim_recv_chan_send_count = 0
                    self._diffs.append(self._gc_data())
                rtn_data = {
                    "aid": "rtn_data",
                    "data": self._diffs,
                }
                self._diffs = []
                self._pending_peek = False
                await self._sim_recv_chan.send(rtn_data)

    async def _generator_diffs(self, keep_current):
        """
        keep_current 为 True 表示不会推进行情，为 False 表示需要推进行情
        即 self._diffs 为 None 并且 keep_current = True 会推进行情
        """
        quotes_helper = {}  # 记录生成行情信息的辅助信息, 用于计算最后返回 quote_diffs
        while self._generators:
            # self._generators 存储了 generator，self._serials 记录一些辅助的信息
            min_request_key = min(self._generators.keys(), key=lambda serial: self._serials[serial]["timestamp"])
            timestamp = self._serials[min_request_key]["timestamp"]  # 所有已订阅数据中的最小行情时间
            is_before_current_dt = timestamp < self._current_dt  # 生成这笔行情的时间是否小于当前回测时间
            # 推进时间，一次只会推进最多一个(补数据时有可能是0个)行情时间，并确保<=该行情时间的行情都被发出
            # 如果行情时间大于当前回测时间 则 判断是否diff中已有数据；否则表明此行情时间的数据未全部保存在diff中，则继续append
            if timestamp > self._current_dt:
                if self._diffs or keep_current:  # 如果diffs中已有数据：退出循环并发送数据给下游api
                    break
                else:
                    self._current_dt = timestamp  # 否则将回测时间更新至最新行情时间
            diff = self._serials[min_request_key]["diff"]
            self._diffs.append(diff)
            # klines 请求，需要记录已经发送 api 的数据
            for symbol in diff.get("klines", {}):
                for dur in diff["klines"][symbol]:
                    for kid in diff["klines"][symbol][dur].get('data', {}):
                        rs = self._sended_to_api.setdefault((symbol, int(dur)), [])
                        kid = int(kid)
                        self._sended_to_api[(symbol, int(dur))] = _rangeset_range_union(rs, (kid, kid + 1))
            # 记录用于生成 quote 的信息
            quotes_helper[min_request_key] = {
                "timestamp": timestamp,
                "kline_or_tick": self._serials[min_request_key]["kline_or_tick"],
                "when": self._serials[min_request_key]["when"],
                "is_before_current_dt": is_before_current_dt,
            }
            await self._fetch_serial(min_request_key)
        if self._had_any_generator and not self._generators and not self._diffs:  # 当无可发送数据时则抛出BacktestFinished例外,包括未订阅任何行情 或 所有已订阅行情的最后一笔行情获取完成
            self._api._print("回测结束")
            self._logger.debug("backtest finished")
            if self._current_dt < self._end_dt:
                self._current_dt = 2145888000000000000  # 一个远大于 end_dt 的日期 20380101
            await self._sim_recv_chan.send({
                "aid": "rtn_data",
                "data": [{"_tqsdk_backtest": self._get_backtest_time()}]
            })
            await self._api._wait_until_idle()
            raise BacktestFinished(self._api) from None
        # 把 quote 对应的 diffs 添加到 self._diffs 中
        self._generator_quotes_diffs(quotes_helper)  # 生成 quotes_diffs

    def _generator_quotes_diffs(self, quotes_helper) -> dict:
        """
        _generator_diffs 之后批量生成 quotes_diffs，支持只有最小周期 Kline 生成 high low 对应行情
        https://shinnytech.atlassian.net/wiki/spaces/~245981717/pages/1527349338/tqsdk+high+low
        """
        for key in quotes_helper.keys():
            ins, dur = key
            symbol = ins.split(",")[0]
            if quotes_helper[key]["is_before_current_dt"] and self._quotes.get(symbol, {}).get("sended_init_quote"):
                # 先订阅 A 合约，再订阅 A 合约日线，那么 A 合约的行情时间会回退: 2021-01-04 09:31:59.999999 -> 2021-01-01 18:00:00.000000
                # 如果当前 timestamp 小于 _current_dt，那么这个 quote_diff 不需要发到下游
                # 如果先订阅 A 合约（有夜盘），时间停留在夜盘开始时间， 再订阅 B 合约（没有夜盘），那么 B 合约的行情（前一天收盘时间）应该发下去，
                # 否则 get_quote(B) 等到收到行情才返回，会直接把时间推进到第二天白盘。
                continue
            item = quotes_helper[key]["kline_or_tick"]
            if item is None:
                # item 如果是 None 的话，没有可以生成行情的信息的数据，那么不生成 quote_diff
                continue
            diffs = None
            if self._quotes[symbol]['min_duration'] == 0 and dur == 0:
                # tick 生成行情
                diffs = TqBacktest._get_quote_diffs_from_tick(symbol, item)
            if self._quotes[symbol]['min_duration'] != 0:
                # kline 生成行情
                when = quotes_helper[key]["when"]
                timestamp = quotes_helper[key]["timestamp"]  # quote 行情时间
                quote_info = self._data["quotes"][symbol]
                froms = ["open"] if when == "OPEN" else ["close"]
                if when == "CLOSE" and self._quotes[symbol]['min_duration'] == dur:
                    # kline 生成 quote 数据，只有该合约订阅的最小周期会生成 high low 对应的行情
                    froms = ["high", "low", "close"]
                diffs = TqBacktest._get_quote_diffs_from_kline(symbol, quote_info['price_tick'], timestamp, item, froms)
            if diffs:
                self._quotes[symbol]["sended_init_quote"] = True
                self._diffs.extend(diffs)

    def _get_backtest_time(self) -> dict:
        if self._is_first_send:
            self._is_first_send = False
            return {
                    "start_dt": self._start_dt,
                    "current_dt": self._current_dt,
                    "end_dt": self._end_dt
                }
        else:
            return {
                "current_dt": self._current_dt
            }

    async def _ensure_serial(self, ins, dur, chart_id=None):
        if (ins, dur) not in self._serials:
            symbol = ins.split(',')[0]
            quote = self._quotes.setdefault(symbol, {  # 在此处设置 min_duration: 每次生成K线的时候会自动生成quote, 记录某一合约的最小duration
                "min_duration": dur
            })
            quote["min_duration"] = min(quote["min_duration"], dur)
            self._serials[(ins, dur)] = {
                "chart_id_set": {chart_id} if chart_id else set()  # 记录当前 serial 对应的 chart_id
            }
            self._generators[(ins, dur)] = self._gen_serial(ins, dur)
            self._had_any_generator = True
            await self._fetch_serial((ins, dur))
        elif chart_id:
            self._serials[(ins, dur)]["chart_id_set"].add(chart_id)

    async def _ensure_query(self, pack):
        """一定收到了对应 query 返回的包"""
        query_pack = {"query": pack["query"]}
        if query_pack.items() <= self._data.get("symbols", {}).get(pack["query_id"], {}).items():
            return
        async with TqChan(self._api, last_only=True) as update_chan:
            self._data["_listener"].add(update_chan)
            while not query_pack.items() <= self._data.get("symbols", {}).get(pack["query_id"], {}).items():
                await update_chan.recv()

    async def _ensure_quote(self, symbol):
        # 在接新版合约服务器后，合约信息程序运行过程中查询得到的，这里不再能保证合约一定存在，需要添加 quote 默认值
        quote = _get_obj(self._data, ["quotes", symbol], BtQuote(self._api))
        if math.isnan(quote.get("price_tick")):
            query_pack = _query_for_quote(symbol)
            await self._md_send_chan.send(query_pack)
            async with TqChan(self._api, last_only=True) as update_chan:
                quote["_listener"].add(update_chan)
                while math.isnan(quote.get("price_tick")):
                    await update_chan.recv()
        if symbol not in self._quotes or self._quotes[symbol]["min_duration"] > 60000000000:
            await self._ensure_serial(symbol, 60000000000)

    async def _fetch_serial(self, key):
        s = self._serials[key]
        try:
            s["timestamp"], s["diff"], s["kline_or_tick"], s["when"] = await self._generators[key].__anext__()
        except StopAsyncIteration:
            del self._generators[key]  # 删除一个行情时间超过结束时间的 generator

    async def _gen_serial(self, ins, dur):
        """k线/tick 序列的 async generator, yield 出来的行情数据带有时间戳, 因此 _send_diff 可以据此归并"""
        # 先定位左端点, focus_datetime 是 lower_bound ,这里需要的是 upper_bound
        # 因此将 view_width 和 focus_position 设置成一样，这样 focus_datetime 所对应的 k线刚好位于屏幕外
        # 使用两个长度为 10000 的 chart，去缓存/回收下游需要的数据
        chart_id_a = _generate_uuid("PYSDK_backtest")
        chart_id_b = _generate_uuid("PYSDK_backtest")
        chart_info = {
            "aid": "set_chart",
            "chart_id": chart_id_a,
            "ins_list": ins,
            "duration": dur,
            "view_width": 10000,  # 设为 10000 原因：可满足用户所有的订阅长度，并在backtest中将所有的 相同合约及周期 的K线用同一个serial存储
            "focus_datetime": int(self._current_dt),
            "focus_position": 10000,
        }
        chart_a = _get_obj(self._data, ["charts", chart_id_a])
        chart_b = _get_obj(self._data, ["charts", chart_id_b])
        symbol_list = ins.split(',')
        current_id = None  # 当前数据指针
        if dur == 0:
            serials = [_get_obj(self._data, ["ticks", symbol_list[0]])]
        else:
            serials = [_get_obj(self._data, ["klines", s, str(dur)]) for s in symbol_list]
        async with TqChan(self._api, last_only=True) as update_chan:
            for serial in serials:
                serial["_listener"].add(update_chan)
            chart_a["_listener"].add(update_chan)
            chart_b["_listener"].add(update_chan)
            await self._md_send_chan.send(chart_info.copy())
            try:
                async for _ in update_chan:
                    chart = _get_obj(self._data, ["charts", chart_info["chart_id"]])
                    if not (chart_info.items() <= _get_obj(chart, ["state"]).items()):
                        # 当前请求还没收齐回应, 不应继续处理
                        continue
                    if not chart.get("ready", False):
                        continue  # chart 数据还没准备好
                    last_id = serials[0].get("last_id", -1)
                    if last_id == -1:
                        # 所有合约的 tick 数据一定有，开盘一定会收到一笔 tick
                        # kline 是由有价格的 tick 生成的，所以 kline 可能没有数据的
                        assert dur > 0
                        diff = {
                            "klines": {
                                symbol_list[0]: {
                                    str(dur): {
                                        "last_id": -1
                                    }
                                }
                            }
                        }
                        yield self._current_dt, diff, None, "OPEN"
                        return
                    if self._data.get("mdhis_more_data", True):
                        self._data["_listener"].add(update_chan)
                        continue
                    else:
                        self._data["_listener"].discard(update_chan)
                    left_id = chart.get("left_id", -1)
                    right_id = chart.get("right_id", -1)
                    if current_id is None:
                        current_id = max(left_id, 0)
                    # 发送下一段 chart 10000 根 kline
                    chart_info["chart_id"] = chart_id_b if chart_info["chart_id"] == chart_id_a else chart_id_a
                    chart_info["left_kline_id"] = right_id
                    chart_info.pop("focus_datetime", None)
                    chart_info.pop("focus_position", None)
                    await self._md_send_chan.send(chart_info.copy())
                    while True:
                        if current_id > last_id:
                            # 当前 id 已超过 last_id
                            return
                        # 将订阅的10000长度的窗口中的数据都遍历完后，退出循环，然后再次进入并处理下一窗口数据
                        if current_id > right_id:
                            break
                        item = {k: v for k, v in serials[0]["data"].get(str(current_id), {}).items()}
                        if dur == 0:
                            diff = {
                                "ticks": {
                                    ins: {
                                        "last_id": current_id,
                                        "data": {
                                            str(current_id): item,
                                            str(current_id - 10000): None,
                                        }
                                    }
                                }
                            }
                            if item["datetime"] > self._end_dt:  # 超过结束时间
                                return
                            yield item["datetime"], diff, item, "TICK"
                        else:
                            timestamp = item["datetime"] if dur < 86400000000000 else _get_trading_day_start_time(
                                item["datetime"])
                            if timestamp > self._end_dt:  # 超过结束时间
                                return
                            binding = serials[0].get("binding", {})
                            diff = {
                                "klines": {
                                    symbol_list[0]: {
                                        str(dur): {
                                            "last_id": current_id,
                                            "data": {
                                                str(current_id): {
                                                    "datetime": item["datetime"],
                                                    "open": item["open"],
                                                    "high": item["open"],
                                                    "low": item["open"],
                                                    "close": item["open"],
                                                    "volume": 0,
                                                    "open_oi": item["open_oi"],
                                                    "close_oi": item["open_oi"],
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                            for chart_id in self._serials[(ins, dur)]["chart_id_set"]:
                                diff["charts"] = {
                                    chart_id: {
                                        "right_id": current_id  # api 中处理多合约 kline 需要 right_id 信息
                                    }
                                }
                            for i, symbol in enumerate(symbol_list):
                                if i == 0:
                                    diff_binding = diff["klines"][symbol_list[0]][str(dur)].setdefault("binding", {})
                                    continue
                                other_id = binding.get(symbol, {}).get(str(current_id), -1)
                                if other_id >= 0:
                                    diff_binding[symbol] = {str(current_id): str(other_id)}
                                    other_item = serials[i]["data"].get(str(other_id), {})
                                    diff["klines"][symbol] = {
                                        str(dur): {
                                            "last_id": other_id,
                                            "data": {
                                                str(other_id): {
                                                    "datetime": other_item["datetime"],
                                                    "open": other_item["open"],
                                                    "high": other_item["open"],
                                                    "low": other_item["open"],
                                                    "close": other_item["open"],
                                                    "volume": 0,
                                                    "open_oi": other_item["open_oi"],
                                                    "close_oi": other_item["open_oi"],
                                                }
                                            }
                                        }
                                    }
                            yield timestamp, diff, item, "OPEN"  # K线刚生成时的数据都为开盘价
                            timestamp = item["datetime"] + dur - 1000 \
                                if dur < 86400000000000 else _get_trading_day_start_time(item["datetime"] + dur) - 1000
                            if timestamp > self._end_dt:  # 超过结束时间
                                return
                            diff = {
                                "klines": {
                                    symbol_list[0]: {
                                        str(dur): {
                                            "data": {
                                                str(current_id): item,
                                            }
                                        }
                                    }
                                }
                            }
                            for i, symbol in enumerate(symbol_list):
                                if i == 0:
                                    continue
                                other_id = binding.get(symbol, {}).get(str(current_id), -1)
                                if other_id >= 0:
                                    diff["klines"][symbol] = {
                                        str(dur): {
                                            "data": {
                                                str(other_id): {k: v for k, v in
                                                                serials[i]["data"].get(str(other_id), {}).items()}
                                            }
                                        }
                                    }
                            yield timestamp, diff, item, "CLOSE"  # K线结束时生成quote数据
                        current_id += 1
            finally:
                # 释放chart资源
                chart_info["ins_list"] = ""
                await self._md_send_chan.send(chart_info.copy())
                chart_info["chart_id"] = chart_id_b if chart_info["chart_id"] == chart_id_a else chart_id_a
                await self._md_send_chan.send(chart_info.copy())

    def _gc_data(self):
        # api 应该删除的数据 diff
        need_rangeset = {}
        for ins, dur in self._serials:
            if dur == 0:  # tick 在发送数据过程中已经回收内存
                continue
            symbol_list = ins.split(',')
            for s in symbol_list:
                need_rangeset.setdefault((s, dur), [])
            main_serial = _get_obj(self._data, ["klines", symbol_list[0], str(dur)])
            main_serial_rangeset = self._sended_to_api.get((symbol_list[0], dur), [])  # 此 request 还没有给 api 发送过任何数据时为 []
            if not main_serial_rangeset:
                continue
            last_id = main_serial_rangeset[-1][-1] - 1
            assert last_id > -1
            need_rangeset[(symbol_list[0], dur)] = _rangeset_range_union(need_rangeset[(symbol_list[0], dur)],
                                                                         (last_id - 8963, last_id + 1))
            for symbol in symbol_list[1:]:
                symbol_need_rangeset = []
                symbol_binding = main_serial.get("binding", {}).get(symbol, {})
                if symbol_binding:
                    for i in range(last_id - 8963, last_id + 1):
                        other_id = symbol_binding.get(str(i))
                        if other_id:
                            symbol_need_rangeset = _rangeset_range_union(symbol_need_rangeset, (other_id, other_id + 1))
                if symbol_need_rangeset:
                    need_rangeset[(symbol, dur)] = _rangeset_union(need_rangeset[(symbol, dur)], symbol_need_rangeset)

        gc_rangeset = {}
        for key, rs in self._sended_to_api.items():
            gc_rangeset[key] = _rangeset_difference(rs, need_rangeset.get(key, []))

        # 更新 self._sended_to_api
        for key, rs in gc_rangeset.items():
            self._sended_to_api[key] = _rangeset_difference(self._sended_to_api[key], rs)

        gc_klines_diff = {}
        for (symbol, dur), rs in gc_rangeset.items():
            gc_klines_diff.setdefault(symbol, {})
            gc_klines_diff[symbol][str(dur)] = {"data": {}}
            serial = _get_obj(self._data, ["klines", symbol, str(dur)])
            serial_binding = serial.get("binding", None)
            if serial_binding:
                gc_klines_diff[symbol][str(dur)]["binding"] = {s: {} for s in serial_binding.keys()}
            for start_id, end_id in rs:
                for i in range(start_id, end_id):
                    gc_klines_diff[symbol][str(dur)]["data"][str(i)] = None
                    if serial_binding:
                        for s, s_binding in serial_binding.items():
                            gc_klines_diff[symbol][str(dur)]["binding"][s][str(i)] = None
        return {"klines": gc_klines_diff}

    @staticmethod
    def _get_quote_diffs_from_tick(symbol, tick):
        quote = {k: v for k, v in tick.items()}
        quote["datetime"] = _timestamp_nano_to_str(tick["datetime"])
        return [{
            "quotes": {
                symbol: quote
            }
        }]

    @staticmethod
    def _get_quote_diffs_from_kline(symbol, price_tick, timestamp, kline, froms: list):
        """
        根据 kline 生成 quote
        froms: 列表，每一项为以下 "open", "high", "low", "close" 之一，表示根据哪个价格生成 quote 盘口；
            其中最后一项只能是 open 或 close，表示生成 quote 的最新价

        分为三个包发给下游：
        1. 根据 diff 协议，对于用户收到的最终结果没有影响
        2. TqSim 撮合交易会按顺序处理收到的包，分别比较 high、low、close 三个价格对应的买卖价
        3. TqSim 撮合交易只用到了买卖价，所以最新价只产生一次 close，而不会发送三次
        4. 最高最低仅用于 TqSim 撮合交易，由最小周期生成
        """
        diffs = []
        assert froms[-1] in ["open", "close"]
        for index, key in enumerate(froms):
            if index == 0:
                diffs.append({
                    "quotes": {
                        symbol: {
                            "datetime": _timestamp_nano_to_str(timestamp),
                            "ask_price1": kline[key] + price_tick,
                            "ask_volume1": 1,
                            "bid_price1": kline[key] - price_tick,
                            "bid_volume1": 1,
                            "last_price": kline[froms[-1]],
                            "highest": float("nan"),
                            "lowest": float("nan"),
                            "average": float("nan"),
                            "volume": 0,
                            "amount": float("nan"),
                            "open_interest": kline[f'{froms[-1]}_oi'],
                        }
                    }
                })
            else:
                diffs.append({
                    "quotes": {
                        symbol: {
                            "ask_price1": kline[key] + price_tick,
                            "bid_price1": kline[key] - price_tick,
                        }
                    }
                })
        return diffs



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'mayanqiong'

"""
定义常量
"""

FUTURE_EXCHANGES = ["CFFEX", "SHFE", "DCE", "CZCE", "INE", "GFEX"]
STOCK_EXCHANGES = ["SSE", "SZSE"]
SPOT_EXCHANGES = ["SSWE"]
KQ_EXCHANGES = ["KQ"]
KQD_EXCHANGES = ["KQD"]



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'mayanqiong'


from tqsdk.datetime import _get_trading_day_end_time, _get_trading_day_from_timestamp, _get_expire_rest_days
from tqsdk.datetime_state import TqDatetimeState
from tqsdk.diff import _simple_merge_diff, _merge_diff
from tqsdk.entity import Entity
from tqsdk.utils import _query_for_quote


class CustomDict(Entity):
    """ Position / Order / Trade 对象 """

    def __init__(self, api, new_objs_list):
        self._api = api
        self._new_objs_list = new_objs_list

    def __setitem__(self, key, value):
        self._new_objs_list.append(value)
        return super(CustomDict, self).__setitem__(key, value)


class TradeExtension():
    """
    为持仓、委托单、成交对象添加 合约信息

    * 为期权合约相应的持仓、委托单、成交，添加以下字段
        + option_class 代表期权方向 CALL or PUT，非期权合约该处显示为NONE
        + underlying_symbol
        + strike_price
        + expire_rest_days 距离到期日剩余天数

    """

    def __init__(self, api):
        self._api = api
        self._data = Entity()  # 交易业务信息截面，需要定于数据原型，使用 Entity 类型 和 _merge_diff
        self._data._instance_entity([])
        self._new_objs_list = []
        self._prototype = {
            "trade": {
                "*": {
                    "@": CustomDict(self._api, self._new_objs_list)
                }
            }
        }
        self._data_quotes = {}  # 行情信息截面，只需要 quotes 数据。这里不需要定义数据原型，使用普通 dict 和 _simple_merge_diff
        self._diffs = []
        self._all_trade_symbols = set()  # 所有持仓、委托、成交中的合约
        self._query_symbols = set()  # 已经发送合约信息请求 + 已经知道合约信息的合约
        self._need_wait_symbol_info = set()  # 需要发送合约信息请求 + 不知道合约信息的合约

    async def _run(self, api_send_chan, api_recv_chan, md_send_chan, md_recv_chan):
        self._logger = self._api._logger.getChild("TradeExtension")
        self._api_send_chan = api_send_chan
        self._api_recv_chan = api_recv_chan
        self._md_send_chan = md_send_chan
        self._md_recv_chan = md_recv_chan
        self._datetime_state = TqDatetimeState()
        self._trading_day_end = None
        md_task = self._api.create_task(self._md_handler())
        self._pending_peek = False  # True 表示收到下游的 peek_message ，并且没有发给过下游回复；False 表示发给过下游回复，没有 pending_peek_message
        self._pending_peek_md = False  # True 表示发给过上游 peek_message；False 表示对上游没有 pending_peek_message
        try:
            async for pack in api_send_chan:
                if "_md_recv" in pack:
                    self._pending_peek_md = False
                    await self._md_recv(pack)
                    await self._send_diff()
                    if self._pending_peek and self._pending_peek_md is False:
                        self._pending_peek_md = True
                        await self._md_send_chan.send({"aid": "peek_message"})
                elif pack["aid"] == "peek_message":
                    self._pending_peek = True
                    await self._send_diff()
                    if self._pending_peek and self._pending_peek_md is False:
                        self._pending_peek_md = True
                        await self._md_send_chan.send(pack)
                else:
                    await self._md_send_chan.send(pack)
        finally:
            await self._api._cancel_task(md_task)

    async def _md_handler(self):
        """0 接收上游数据包 """
        async for pack in self._md_recv_chan:
            pack["_md_recv"] = True
            await self._api_send_chan.send(pack)

    async def _md_recv(self, pack):
        """ 处理下行数据包
        0 将行情数据和交易数据合并至 self._data
        1 生成增量业务截面, 该截面包含期权补充的字段
        """
        for d in pack.get("data", {}):
            self._datetime_state.update_state(d)
            _simple_merge_diff(self._data_quotes, d.get('quotes', {}))
            _merge_diff(self._data, {"trade": d.get('trade', {})}, prototype=self._prototype, persist=False, reduce_diff=False)
            self._diffs.append(d)  # 添加至 self._diff 等待被发送

        for obj in self._new_objs_list:
            # 新添加的 Position / Order / Trade  节点
            if hasattr(obj, '_path') and obj['_path'][2] in ['positions', 'trades', 'orders']:
                symbol = f"{obj.get('exchange_id', '')}.{obj.get('instrument_id', '')}"
                if symbol not in self._all_trade_symbols:
                    self._all_trade_symbols.add(symbol)
                    self._need_wait_symbol_info.add(symbol)  # 需要发送合约信息请求

        for s in self._need_wait_symbol_info.copy():
            if self._data_quotes.get(s, {}).get("price_tick", 0) > 0:
                self._need_wait_symbol_info.remove(s)  # 需要发送合约信息请求 + 不知道合约信息的合约

        # 不知道合约信息 并且未发送请求查询合约信息
        unknown_symbols = self._need_wait_symbol_info - self._query_symbols
        if len(unknown_symbols) > 0:
            self._query_symbols = self._query_symbols.union(unknown_symbols)  # 所有发送过ins_query的合约
            query_pack = _query_for_quote(list(unknown_symbols))
            await self._md_send_chan.send(query_pack)

    def _generate_pend_diff(self):
        """"
        补充期权额外字段
        此函数在 send_diff() 才会调用， self._datetime_state.data_ready 一定为 True，
        调用 self._datetime_state.get_current_dt() 一定有正确的当前时间
        """
        pend_diff = {}
        account_keys = list(self._data.get('trade', {}).keys())
        objs_keys = ['positions', 'trades', 'orders']

        # 如果有新添加的合约, 只填充一次即可
        if self._new_objs_list:
            pend_diff.setdefault('trade', {k: {o_k: {} for o_k in objs_keys} for k in account_keys})
            for obj in self._new_objs_list:
                # 新添加的 Position / Order / Trade  节点
                if hasattr(obj, '_path') and obj['_path'][2] in objs_keys:
                    account_key = obj['_path'][1]
                    obj_key = obj['_path'][2]
                    item_id = obj['_path'][3]
                    quote = self._data_quotes.get(f"{obj.get('exchange_id', '')}.{obj.get('instrument_id', '')}", {})
                    if quote.get('ins_class', '').endswith('OPTION'):
                        pend_diff_item = pend_diff['trade'][account_key][obj_key].setdefault(item_id, {})
                        pend_diff_item['option_class'] = quote.get('option_class')
                        pend_diff_item['strike_price'] = quote.get('strike_price')
                        pend_diff_item['underlying_symbol'] = quote.get('underlying_symbol')
                        if quote.get('expire_datetime'):
                            pend_diff_item['expire_rest_days'] = _get_expire_rest_days(quote.get('expire_datetime'),
                                                                                       self._datetime_state.get_current_dt() / 1e9)
            self._new_objs_list.clear()

        # 如果有切换交易日，所有合约都需要修改 expire_rest_days
        current_dt = self._datetime_state.get_current_dt()
        if self._trading_day_end is None or current_dt > self._trading_day_end:
            pend_diff.setdefault('trade', {k: {o_k: {} for o_k in objs_keys} for k in account_keys})
            for account_key, account_node in self._data.get('trade', {}).items():
                for k in objs_keys:
                    for item_id, item in account_node.get(k, {}).items():
                        quote = self._data_quotes.get(f"{item['exchange_id']}.{item['instrument_id']}", {})
                        if quote.get('ins_class', '').endswith('OPTION') and quote.get('expire_datetime'):
                            pend_diff_item = pend_diff['trade'][account_key][k].setdefault(item_id, {})
                            # 剩余到期日字段，每天都会更新，每次都重新计算
                            pend_diff_item['expire_rest_days'] = _get_expire_rest_days(quote.get('expire_datetime'),
                                                                                       current_dt / 1e9)
            self._trading_day_end = _get_trading_day_end_time(_get_trading_day_from_timestamp(current_dt))
        return pend_diff

    async def _send_diff(self):
        if self._datetime_state.data_ready and self._pending_peek and self._diffs and len(self._need_wait_symbol_info) == 0:
            # 生成增量业务截面, 该截面包含期权补充的字段，只在真正需要给下游发送数据时，才将需要发送的数据放在 _diffs 中
            pend_diff = self._generate_pend_diff()
            self._diffs.append(pend_diff)
            rtn_data = {
                "aid": "rtn_data",
                "data": self._diffs,
            }
            self._diffs = []
            self._pending_peek = False
            await self._api_recv_chan.send(rtn_data)



#!usr/bin/env python3
# -*- coding:utf-8 -*-
__author__ = 'yanqiong'

import hashlib
import base64
from urllib.parse import urlparse
from typing import Optional

from shinny_structlog import ShinnyLoggerAdapter

from tqsdk.tradeable.mixin import FutureMixin, StockMixin
from tqsdk.tradeable.tradeable import Tradeable
from tqsdk.channel import TqChan
from tqsdk.connect import TqConnect, TdReconnectHandler


class BaseOtg(Tradeable):
    def __init__(self, broker_id: str, account_id: str, password: str, td_url: Optional[str] = None, sm: bool = False) -> None:
        if not isinstance(broker_id, str):
            raise Exception("broker_id 参数类型应该是 str")
        if not isinstance(account_id, str):
            raise Exception("account_id 参数类型应该是 str")
        if not isinstance(password, str):
            raise Exception("password 参数类型应该是 str")
        self._broker_id = broker_id.strip()  # 期货公司（用户登录 rsp_login 填的）
        self._account_id = account_id.strip()  # 期货账户 （用户登录 rsp_login 填的）
        self._password = password
        self._td_url = td_url
        self._sm = sm

        super(BaseOtg, self).__init__()

    def _get_account_key(self):
        s = self._broker_id + self._account_id
        return hashlib.md5(s.encode('utf-8')).hexdigest()

    @property
    def _account_name(self):
        return self._account_id

    @property
    def _account_info(self):
        info = super(BaseOtg, self)._account_info
        info.update({
            "broker_id": self._broker_id,
            "account_id": self._account_id
        })
        return info

    async def _send_login_pack(self):
        """发送登录请求"""
        req = {
            "aid": "req_login",
            "bid": self._broker_id,
            "user_name": self._account_id,
            "password": self._password
        }
        await self._td_send_chan.send(req)

    def _update_otg_info(self, api):
        """更新 otg 登录需要的基本信息"""
        if self._td_url:
            return
        if api._td_url:
            self._td_url = api._td_url
        else:
            self._td_url, account_type, sm_type, sm_config = api._auth._get_td_url(self._broker_id, self._account_id)
            if account_type == "FUTURE":
                assert isinstance(self, FutureMixin)
            else:
                assert isinstance(self, StockMixin)
            if self._sm and sm_type and sm_config:
                url_account = base64.urlsafe_b64encode(self._account_id.encode("utf-8")).decode("utf-8")
                url_password = base64.urlsafe_b64encode(self._password.encode("utf-8")).decode("utf-8")
                url_info = urlparse(self._td_url)
                # http://example.org -> http://example.org/smcfg/smuser/smpasswd
                # http://example.org/ -> http://example.org/smcfg/smuser/smpasswd/
                # http://example.org/foo/bar -> http://example.org/smcfg/smuser/smpasswd/foo/bar
                self._td_url = url_info._replace(scheme=sm_type, path=f"/{sm_config}/{url_account}/{url_password}{url_info.path}").geturl()

    def _connect_td(self, api, index: int) -> Optional[str]:
        # 连接交易服务器
        td_logger = ShinnyLoggerAdapter(api._logger.getChild("TqConnect"), url=self._td_url, broker_id=self._broker_id, account_id=self._account_id)
        conn_id = f"td_{index}"
        ws_td_send_chan = TqChan(api, chan_name=f"send to {conn_id}", logger=td_logger)
        ws_td_recv_chan = TqChan(api, chan_name=f"recv from {conn_id}", logger=td_logger)
        conn = TqConnect(td_logger, conn_id=conn_id)
        api.create_task(conn._run(api, self._td_url, ws_td_send_chan, ws_td_recv_chan))
        ws_td_send_chan._logger_bind(chan_from=f"td_reconn_{index}")
        ws_td_recv_chan._logger_bind(chan_to=f"td_reconn_{index}")

        td_handler_logger = ShinnyLoggerAdapter(api._logger.getChild("TdReconnect"), url=self._td_url, broker_id=self._broker_id, account_id=self._account_id)
        td_reconnect = TdReconnectHandler(td_handler_logger)
        self._td_send_chan = TqChan(api, chan_name=f"send to td_reconn_{index}", logger=td_handler_logger)
        self._td_recv_chan = TqChan(api, chan_name=f"recv from td_reconn_{index}", logger=td_handler_logger)
        api.create_task(td_reconnect._run(api, self._td_send_chan, self._td_recv_chan, ws_td_send_chan, ws_td_recv_chan))
        self._td_send_chan._logger_bind(chan_from=f"account_{index}")
        self._td_recv_chan._logger_bind(chan_to=f"account_{index}")
        return conn_id

    async def _run(self, api, api_send_chan, api_recv_chan, md_send_chan, md_recv_chan):
        self._api = api
        self._api_send_chan = api_send_chan
        self._api_recv_chan = api_recv_chan
        self._md_send_chan = md_send_chan
        self._md_recv_chan = md_recv_chan
        await self._send_login_pack()
        await super(BaseOtg, self)._run(api, api_send_chan, api_recv_chan, md_send_chan, md_recv_chan, self._td_send_chan, self._td_recv_chan)

    async def _handle_recv_data(self, pack, chan):
        """
        处理所有上游收到的数据包
        """
        if chan == self._md_recv_chan:  # 从行情收到的数据包
            if pack["aid"] == "rtn_data":
                self._diffs.extend(pack.get('data', []))
            else:
                await self._api_recv_chan.send(pack)  # 有可能是另一个 account 的 rsp_login
        elif chan == self._td_recv_chan:  # 从交易收到的数据包
            # 收到通知时，在通知里加上 account_name 信息
            if pack["aid"] == "rtn_data":
                for data in pack.get('data', []):
                    for notify_id, notify in data.get('notify', {}).items():
                        notify['_account_name'] = self._account_name
            self._td_handler(pack)

    async def _handle_req_data(self, pack):
        if self._is_self_trade_pack(pack):
            await self._td_send_chan.send(pack)
        else:
            await self._md_send_chan.send(pack)

    def _td_handler(self, pack):
        # OTG 返回业务信息截面 trade 中 account_key 为 user_id, 该值需要替换为 account_key
        if pack["aid"] == "rtn_data":
            pack_data = pack.get('data', [])
            for item in pack_data:
                if "trade" in item:
                    item["trade"][self._account_key] = item["trade"].pop(self._account_id)
            self._diffs.extend(pack_data)



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'mayanqiong'

from tqsdk.datetime import _get_trading_day_from_timestamp, _get_trade_timestamp, _timestamp_nano_to_str

TRADING_DAYS_OF_YEAR = 250


def _get_price_range(quote):
    """ 返回合约对应的买一卖一盘口价格"""
    ask_price = quote["ask_price1"]
    bid_price = quote["bid_price1"]
    if quote["ins_class"].endswith("INDEX"):
        # 在指数交易时，使用 tick 进行回测时，backtest 发的 quote 没有买一卖一价；或者在实时行情中，指数的 quote 也没有买一卖一价
        if ask_price != ask_price:
            ask_price = quote["last_price"] + quote["price_tick"]
        if bid_price != bid_price:
            bid_price = quote["last_price"] - quote["price_tick"]
    return ask_price, bid_price


def _get_option_margin(quote, last_price, underlying_last_price):
    """返回每张期权占用保证金，只有空头持仓占用保证金"""
    # 期权保证金计算公式参考深交所文档 http://docs.static.szse.cn/www/disclosure/notice/general/W020191207433397366259.pdf
    if quote["option_class"] == "CALL":
        # 认购期权义务仓开仓保证金＝[合约最新价 + Max（12% × 合约标的最新价 - 认购期权虚值， 7% × 合约标的最新价）] × 合约单位
        # 认购期权虚值＝Max（行权价 - 合约标的最新价，0）
        out_value = max(quote["strike_price"] - underlying_last_price, 0)
        return (last_price + max(0.12 * underlying_last_price - out_value,
                                 0.07 * underlying_last_price)) * quote["volume_multiple"]
    else:
        # 认沽期权义务仓开仓保证金＝Min[合约最新价+ Max（12% × 合约标的最新价 - 认沽期权虚值，7% × 行权价），行权价] × 合约单位
        # 认沽期权虚值＝Max（合约标的最新价 - 行权价，0）
        out_value = max(underlying_last_price - quote["strike_price"], 0)
        return min(quote["last_price"] + max(0.12 * underlying_last_price - out_value,
                                             0.07 * quote["strike_price"]),
                   quote["strike_price"]) * quote["volume_multiple"]


def _get_premium(trade, quote):
    """返回成交导致的权利金变化"""
    if quote["ins_class"].endswith("OPTION"):
        premium = trade["price"] * trade["volume"] * quote["volume_multiple"]
        return -premium if trade["direction"] == "BUY" else premium
    else:
        return 0


def _get_close_profit(trade, quote, position):
    """返回成交导致的平仓盈亏变化"""
    if quote["ins_class"].endswith("OPTION"):
        # 期权没有持仓盈亏没有持仓价，其平仓的盈亏体现在市价变化中
        return 0
    # 期货及其他使用持仓价计算
    elif trade["direction"] == "SELL":
        return (trade["price"] - position["position_price_long"]) * trade["volume"] * quote["volume_multiple"]
    else:
        return (position["position_price_short"] - trade["price"]) * trade["volume"] * quote["volume_multiple"]


def _get_commission(quote={}):
    """返回每手手续费"""
    if quote.get("ins_class", "").endswith("OPTION"):
        return quote.get("user_commission", 10)  # 期权quote没有commission字段, 设为固定10元一张, 优先采用用户设置的参数
    return quote.get("user_commission", quote.get('commission', float('nan')))


def _get_future_margin(quote={}):
    """返回期货每手保证金"""
    if quote.get("ins_class", "").endswith("OPTION"):
        return float('nan')
    return quote.get("user_margin", quote.get("margin", float('nan')))


def _get_order_price(quote, order):
    # order 预期成交价格
    if order["price_type"] in ["ANY", "BEST", "FIVELEVEL"]:
        ask_price, bid_price = _get_price_range(quote)
        return ask_price if order["direction"] == "BUY" else bid_price
    else:
        return order["limit_price"]


def _get_stock_fee(direction, volume, price):
    # 费用(BUY) = 佣金; 费用(SELL) = 佣金 + 印花税
    balance = volume * price
    return max(balance * 0.00025, 5.0) + (0 if direction == "BUY" else balance * 0.001)


def _get_dividend_ratio(quote):
    # 获取合约下一个交易日的送股、分红信息
    timestamp = _get_trading_day_from_timestamp(_get_trade_timestamp(quote['datetime'], float('nan')) + 86400000000000)  # 下一交易日
    stock_dividend = _get_dividend_ratio_by_dt(quote['stock_dividend_ratio'], timestamp=timestamp)
    cash_dividend = _get_dividend_ratio_by_dt(quote['cash_dividend_ratio'], timestamp=timestamp)
    return stock_dividend, cash_dividend


def _get_dividend_ratio_by_dt(dividend_list: list, timestamp: int) -> float:
    # 从分红/送股列表中找到指定的数据返回
    # ['20181102,0.400000', '20200624,0.400000', '20210716,0.400000']  '20210716'
    dt = _timestamp_nano_to_str(timestamp, fmt='%Y%m%d')
    for item in dividend_list:
        if item[:8] == dt:
            return float(item[9:])
    return 0.0



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'mayanqiong'


from pandas import DataFrame


class TqTimeTable(DataFrame):

    def __init__(self, account=None):
        self.__dict__["_account"] = account
        self.__dict__["_columns"] = ['interval', 'volume', 'price']
        super(TqTimeTable, self).__init__(data=[], columns=self.__dict__["_columns"])



#!usr/bin/env python3
# -*- coding:utf-8 -*-

__author__ = 'mayanqiong'

import asyncio
from abc import abstractmethod


class TqModule(object):

    async def _run(self, api, api_send_chan, api_recv_chan, *args):
        """
        可以接入 n 个上游，1 个下游
        """
        assert len(args) % 2 == 0 and len(args) > 0
        self._diffs = []
        self._up_chans = [{"send_chan": args[i], "recv_chan": args[i + 1]} for i in range(0, len(args), 2)]
        self._up_pending_peek = [False for _ in self._up_chans]  # 是否有发给上游的 peek_message，未收到过回复
        self._pending_peek = False  # 是否有下游收到未处理的 peek_message
        up_handle_tasks = [
            api.create_task(self._up_handler(api_send_chan, item["recv_chan"], chan_index=i))
            for i, item in enumerate(self._up_chans)
            ]
        try:
            async for pack in api_send_chan:
                if pack["aid"] == "peek_message":
                    # 处理下游发送的 peek_message
                    self._pending_peek = True
                    await self._send_diff(api_recv_chan)
                    if self._pending_peek:
                        for i, pending_peek in enumerate(self._up_pending_peek):
                            if pending_peek is False:
                                # 控制"peek_message"发送: 下游发送了 peek_message，并且上游没有 pending_peek 才会发送
                                await self._up_chans[i]["send_chan"].send({"aid": "peek_message"})
                                self._up_pending_peek[i] = True
                elif pack.get('_up_chan_index', None) is not None:
                    # 处理上游发送的数据包
                    _up_chan_index = pack.pop("_up_chan_index")
                    await self._handle_recv_data(pack, self._up_chans[_up_chan_index]["recv_chan"])
                    await self._send_diff(api_recv_chan)
                else:
                    # 处理下游发送的其他请求
                    await self._handle_req_data(pack)
                    await self._send_diff(api_recv_chan)
        finally:
            await self._api._cancel_tasks(*up_handle_tasks)

    async def _up_handler(self, api_send_chan, recv_chan, chan_index):
        async for pack in recv_chan:
            pack['_up_chan_index'] = chan_index
            if pack['aid'] == 'rtn_data':
                self._up_pending_peek[chan_index] = False
            await api_send_chan.send(pack)

    async def _send_diff(self, api_recv_chan):
        pk = self._pending_peek
        if self._pending_peek and self._diffs:
            rtn_data = {
                "aid": "rtn_data",
                "data": self._diffs,
            }
            self._diffs = []
            self._pending_peek = False
            await api_recv_chan.send(rtn_data)
        await self._on_send_diff(pk)

    @abstractmethod
    async def _handle_recv_data(self, pack, chan):
        """
        处理所有上游收到的数据包，这里应该将需要发送给下游的数据 append 到 self._diffs
        pack: 收到的数据包
        chan: 收到此数据包的 channel
        """
        pass

    @abstractmethod
    async def _handle_req_data(self, pack):
        """
        处理所有下游发送的非 peek_message 数据包
        这里应该将发送的请求转发到指定的某个上游 channel
        """
        pass

    async def _on_send_diff(self, peeding_pk):
        """
        TqModule 调用 _send_diff 时，最后会执行的回调函数；
        TqModule 处理所有 pack 都会调用 self._send_diff()，可以在此处处理额外业务逻辑。
        pending_peek 表示调用 _send_diff 时，self._pending_peek 的值
        """
        pass



#!usr/bin/env python3
# -*- coding:utf-8 -*-
__author__ = 'yanqiong'

import hashlib

from tqsdk.tradeable.otg.base_otg import BaseOtg
from tqsdk.tradeable.mixin import FutureMixin


class TqZq(BaseOtg, FutureMixin):
    """众期账户类"""

    def __init__(self, account_id: str, password: str, td_url: str) -> None:
        """
        创建众期账户实例

        Args:
            account_id (str): 帐号

            password (str): 密码

            td_url (str): 众期交易服务器地址, eg: "ws://1.2.3.4:8765/"

        Example1::

            from tqsdk import TqApi, TqZq
            account = TqZq(account_id="众期账户", password="众期密码", td_url="众期柜台地址")
            api = TqApi(account, auth=TqAuth("快期账户", "账户密码"))

        """
        super(TqZq, self).__init__(broker_id="", account_id=account_id, password=password, td_url=td_url)

    def _get_account_key(self):
        s = self._account_id + self._td_url
        return hashlib.md5(s.encode('utf-8')).hexdigest()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
__time__ = '2020/8/5 22:45'
__author__ = 'Hong Yan'

from typing import List, Union, Optional

from shinny_structlog import ShinnyLoggerAdapter

from tqsdk.channel import TqChan
from tqsdk.tradeable import TqAccount, TqKq, TqKqStock, TqSim, TqSimStock, BaseSim, TqZq, TqCtp
from tqsdk.tradeable.mixin import StockMixin


class TqMultiAccount(object):
    """
    天勤多账户 - TqMultiAccount

    天勤多账户模块提供了单 `api` 同时操作不同账户及其组合的功能支持，目前已支持实盘账户、模拟账户和快期模拟账户的任意组合。
    使用天勤多账户进行跨市场或跨账户交易时，可以在不引入多进程和多线程的前提下, 比较方便的传递账户信息进行策略编写,
    同时, 也更方便对不同账户的交易数据进行统计分析。

    **注意**

    - 多账户模式下, 对于 get_position，account，insert_order，set_target_volume 等函数必须指定 account 参数
    - 多账户模式下, 实盘账户的数量受限于快期账户支持实盘账户数, 详见:`更多的实盘交易账户数 <https://doc.shinnytech.com/tqsdk/latest/profession.html#id2>`_

    """

    def __init__(self, accounts: Optional[List[Union[TqAccount, TqKq, TqZq, TqKqStock, TqSim, TqSimStock, TqZq, TqCtp]]] = None):
        """
        创建 TqMultiAccount 实例

        Args:
            accounts (List[Union[TqAccount, TqKq, TqKqStock, TqSim, TqSimStock, TqZq, TqCtp]]): [可选] 多账户列表, 若未指定任何账户, 则为 [TqSim()]

        Example1::

            from tqsdk import TqApi, TqAccount, TqMultiAccount

            account1 = TqAccount("H海通期货", "123456", "123456")
            account2 = TqAccount("H宏源期货", "654321", "123456")
            api = TqApi(TqMultiAccount([account1, account2]), auth=TqAuth("快期账户", "账户密码"))
            # 分别获取账户资金信息
            order1 = api.insert_order(symbol="DCE.m2101", direction="BUY", offset="OPEN", volume=3, account=account1)
            order2 = api.insert_order(symbol="SHFE.au2012C308", direction="BUY", offset="OPEN", volume=3, limit_price=78.0, account=account2)
            while order1.status != "FINISHED" or order2.status != "FINISHED":
                 api.wait_update()
            # 分别获取账户资金信息
            account_info1 = account1.get_account()
            account_info2 = account2.get_account()
            api.close()

        Example2::

            # 多账户模式下使用 TargetPosTask
            from tqsdk import TqApi, TqAccount, TqMultiAccount, TqAuth, TargetPosTask

            account1 = TqAccount("H海通期货", "123456", "123456")
            account2 = TqAccount("H宏源期货", "654321", "123456")
            api = TqApi(TqMultiAccount([account1, account2]), auth=TqAuth("快期账户", "账户密码"))
            symbol1 = "DCE.m2105"
            symbol2 = "DCE.i2101"
            position1 = account1.get_position(symbol1)
            position2 = account2.get_position(symbol2)
            # 多账户模式下, 调仓工具需要指定账户实例
            target_pos1 = TargetPosTask(api, symbol1, account=account1)
            target_pos2 = TargetPosTask(api, symbol2, account=account2)
            target_pos1.set_target_volume(30)
            target_pos2.set_target_volume(80)
            while position1.volume_long != 30 or position2.volume_long != 80:
                api.wait_update()

            api.close()

        """
        self._account_list = accounts if accounts else [TqSim()]
        self._all_sim_account = all([isinstance(a, BaseSim) for a in self._account_list])  # 是否全部为本地模拟帐号
        self._map_conn_id = {}  # 每次建立连接时，记录每个 conn_id 对应的账户
        if self._has_duplicate_account():
            raise Exception("多账户列表中不允许使用重复的账户实例.")

    def _has_duplicate_account(self):
        # 存在相同的账户实例
        account_set = set([a._account_key for a in self._account_list])
        return len(account_set) != len(self._account_list)

    def _check_valid(self, account: Union[str, TqAccount, TqKq, TqKqStock, TqSim, TqSimStock, None]):
        """
        查询委托、成交、资产、委托时, 需要指定账户实例
        account: 类型 str 表示 account_key，其他为账户类型或者 None
        """
        if isinstance(account, str):
            selected_list = [a for a in self._account_list if a._account_key == account]
            return selected_list[0] if selected_list else None
        elif account is None:
            return self._account_list[0] if len(self._account_list) == 1 else None
        else:
            return account if account in self._account_list else None

    def _get_account_id(self, account):
        """ 获取指定账户实例的账户属性 """
        acc = self._check_valid(account)
        return acc._account_id if acc else None

    def _get_account_key(self, account):
        """ 获取指定账户实例的账户属性 """
        acc = self._check_valid(account)
        return acc._account_key if acc else None

    def _is_stock_type(self, account_or_account_key):
        """ 判断账户类型是否为股票账户 """
        acc = self._check_valid(account_or_account_key)
        return isinstance(acc, StockMixin)

    def _get_trade_more_data(self, data):
        """ 获取业务信息截面 trade_more_data 标识，当且仅当所有账户的标识置为 false 时，业务信息截面就绪 """
        trade_more_datas = []
        for account in self._account_list:
            trade_node = data.get("trade", {}).get(account._account_key, {})
            trade_more_data = trade_node.get("trade_more_data", True)
            trade_more_datas.append(trade_more_data)
        return any(trade_more_datas)

    def _setup_connection(self, api, api_send_chan, api_recv_chan, ws_md_send_chan, ws_md_recv_chan):
        self._api = api
        # 将多个 account chain 起来
        log = ShinnyLoggerAdapter(self._api._logger.getChild("TqMultiAccount"))
        for index, account in enumerate(self._account_list):
            _send_chan = api_send_chan if index == len(self._account_list) - 1 else TqChan(self._api, logger=log)
            _recv_chan = api_recv_chan if index == len(self._account_list) - 1 else TqChan(self._api, logger=log)
            _send_chan._logger_bind(chan_name=f"send to account_{index}")
            _recv_chan._logger_bind(chan_name=f"recv from account_{index}")
            ws_md_send_chan._logger_bind(chan_from=f"account_{index}")
            ws_md_recv_chan._logger_bind(chan_to=f"account_{index}")
            conn_id = account._connect_td(self._api, index)
            if conn_id:
                self._map_conn_id[conn_id] = account
            # 启动账户实例
            self._api.create_task(account._run(self._api, _send_chan, _recv_chan, ws_md_send_chan, ws_md_recv_chan))
            ws_md_send_chan, ws_md_recv_chan = _send_chan, _recv_chan



from tqsdk import TqApi, TqAuth

'''
根据输入的ETF期权来查询该期权的交易所规则下的理论卖方保证金，实际情况请以期货公司收取的一手保证金为准
'''

def etf_margin_cal(symbol):
    quote_etf = api.get_quote(symbol)
    # 判断期权标的是不是ETF
    if quote_etf.underlying_symbol in ["SSE.510050", "SSE.510300", "SZSE.159919"]:
        if quote_etf.option_class == "CALL":
            # 认购期权虚值＝Max（行权价-合约标的前收盘价，0）
            call_out_value = max(quote_etf.strike_price - quote_etf.underlying_quote.pre_close, 0)
            # 认购期权义务仓开仓保证金＝[合约前结算价+Max（12%×合约标的前收盘价-认购期权虚值，7%×合约标的前收盘价）]×合约单位
            call_margin = (quote_etf.pre_settlement + max(0.12 * quote_etf.underlying_quote.pre_close - call_out_value,
                                                          0.07 * quote_etf.underlying_quote.pre_close)) * quote_etf.volume_multiple
            return round(call_margin, 2)
        elif quote_etf.option_class == "PUT":
            # 认沽期权虚值＝Max（合约标的前收盘价-行权价，0）
            put_out_value = max(quote_etf.underlying_quote.pre_close - quote_etf.strike_price, 0)
            # 认沽期权义务仓开仓保证金＝Min[合约前结算价+Max（12%×合约标的前收盘价-认沽期权虚值，7%×行权价），行权价]×合约单位。
            put_margin = min(quote_etf.pre_settlement + max(0.12 * quote_etf.underlying_quote.pre_close - put_out_value,
                                                            0.07 * quote_etf.strike_price),
                             quote_etf.strike_price) * quote_etf.volume_multiple
            return round(put_margin, 2)
    else:
        print("输入的不是ETF期权合约")
        return None


# 创建api
api = TqApi(auth=TqAuth("快期账户", "账户密码"))

# 深交所300etf期权
symbol = "SZSE.90000833"

print(etf_margin_cal(symbol))

api.close()



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'chenli'

import asyncio
import json
import os
import tempfile
import signal
import sys
import subprocess
import contextlib
from pathlib import Path
from asyncio.subprocess import DEVNULL

from tqsdk_zq_otg import get_zq_otg_path


class ZqOtgContext(object):
    def __init__(self):
        self._zq_otg_path = get_zq_otg_path()
        self._zq_otg_exe = str(Path(self._zq_otg_path) / "otg_adapter")
        self._zq_otg_env = os.environ.copy()
        self._zq_otg_env["LD_LIBRARY_PATH"] = str(self._zq_otg_path)
        self._zq_otg_proc = None

    async def __aenter__(self):
        self._zq_otg_data_dir = tempfile.TemporaryDirectory()
        self._zq_otg_data_path = Path(self._zq_otg_data_dir.name)
        return self

    async def get_addr(self):
        """无法启动时返回空字符串"""
        # port_file 是创建在 log_file_path 下的
        port_file = self._zq_otg_data_path / "port.json"

        parameters = json.dumps({
            "log_file_path": str(self._zq_otg_data_path),
            "user_file_path": str(self._zq_otg_data_path),
            "host": "127.0.0.1",
            "port": 0,
        })

        if self._zq_otg_proc is not None and sys.platform.startswith("win"):
            # subprocess.Popen 需要调用 poll 才会更新 returncode
            self._zq_otg_proc.poll()
        if self._zq_otg_proc is None or self._zq_otg_proc.returncode is not None:
            if sys.platform.startswith("win"):
                self._zq_otg_proc = subprocess.Popen([self._zq_otg_exe, f"--config={parameters}", "--mode=cmd"], stdin=DEVNULL, stdout=DEVNULL, stderr=DEVNULL, env=self._zq_otg_env)
            else:
                self._zq_otg_proc = await asyncio.create_subprocess_exec(self._zq_otg_exe, f"--config={parameters}", "--mode=cmd", stdin=DEVNULL, stdout=DEVNULL, stderr=DEVNULL, env=self._zq_otg_env)

            for i in range(30):
                if port_file.exists():
                    with open(port_file, 'r') as file:
                        port = json.load(file)["port"]
                        if port != 0:
                            return f"127.0.0.1:{port}"
                await asyncio.sleep(1)

        return ""

    async def __aexit__(self, exc_type, exc, tb):
        if self._zq_otg_proc is not None:
            with contextlib.suppress(ProcessLookupError):
                self._zq_otg_proc.send_signal(signal.CTRL_BREAK_EVENT if sys.platform.startswith("win") else signal.SIGTERM)
            if sys.platform.startswith("win"):
                self._zq_otg_proc.wait()
            else:
                await self._zq_otg_proc.wait()
        self._zq_otg_data_dir.cleanup()



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'chengzhi'

import sys


class BacktestFinished(Exception):
    """
    回测结束会抛出此例外

    Example::

        from datetime import date
        from tqsdk import TqApi, TqAuth, TqBacktest, TargetPosTask, BacktestFinished, TqSim

        sim = TqSim()
        api = TqApi(account=sim,
                    backtest=TqBacktest(start_dt=date(2018, 5, 1), end_dt=date(2018, 6, 1)),
                    auth=TqAuth("快期账户", "账户密码"))

        klines = api.get_kline_serial("DCE.m1901", 5 * 60, data_length=15)  # 获得 m1901 5分钟K线的引用
        target_pos = TargetPosTask(api, "DCE.m1901")

        try:
            while True:  # 策略代码
                api.wait_update()
                if api.is_changing(klines):
                    ma = sum(klines.close.iloc[-15:]) / 15
                    if klines.close.iloc[-1] > ma:
                        target_pos.set_target_volume(5)
                    elif klines.close.iloc[-1] < ma:
                        target_pos.set_target_volume(0)
        except BacktestFinished:
            api.close()
            print("回测结束")
            print(sim.tqsdk_stat)  # 回测时间内账户交易信息统计结果

    """

    _orig_excepthook = None

    def __init__(self, api):
        message = "回测结束"
        super().__init__(message)
        if BacktestFinished._orig_excepthook is None:
            BacktestFinished._orig_excepthook = sys.excepthook

            def _except_catcher(type, value, traceback):
                if type is BacktestFinished:
                    if api._web_gui:
                        try:
                            api._print("----------- Backtest finished, press [Ctrl + C] to exit. -----------")
                            while True:
                                api.wait_update()
                        except KeyboardInterrupt:
                            pass

                    if not api._loop.is_closed():
                        api.close()
                    sys.exit()
                BacktestFinished._orig_excepthook(type, value, traceback)

            sys.excepthook = _except_catcher


class TqTimeoutError(Exception):
    """
    获取数据超时会抛出此例外

    Example::

        from tqsdk import TqApi, TqAuth, TqTimeoutError

        api = TqApi(auth=TqAuth("快期账户", "账户密码"))

        symbols = ["CZCE.RS808", "CZCE.RS809", "CZCE.RS908", "CZCE.RS909"]  # CZCE.RS808 没有任何成交数据，无法取到 kline
        klines = {}

        for s in symbols:
            try:
                klines[s] = api.get_kline_serial(s, 5 * 60, data_length=15)
            except TqTimeoutError as e:
                print(f"获取 {s} 合约K线 超时！！！")

        for s in klines:  # 打印出成功取到的 kline
            print(s, klines[s].iloc[-1])

        while True:
            api.wait_update()
            # 策略代码 。。。。

    """

    def __init__(self, message):
        super().__init__(message)
        self.message = message


class TqBacktestPermissionError(Exception):
    """
    没有回测权限报错
    """

    def __init__(self, message):
        super().__init__(message)
        self.message = message


class TqRiskRuleError(Exception):
    """
    风控触发的报错
    """

    def __init__(self, message):
        super().__init__(message)
        self.message = message



#!usr/bin/env python3
# -*- coding:utf-8 -*-

__author__ = 'mayanqiong'

from tqsdk.tradeable.otg.tqaccount import TqAccount
from tqsdk.tradeable.otg.tqzq import TqZq
from tqsdk.tradeable.otg.tqkq import TqKq, TqKqStock
from tqsdk.tradeable.otg.tqctp import TqCtp



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'chengzhi'

import asyncio
import json
import os
import logging
import uuid
import sys
import subprocess
from pathlib import Path
from asyncio.subprocess import PIPE

from tqsdk_sm import get_sm_path


class SMContext(object):
    def __init__(self, logger, api, sm_type, sm_data, sm_user, sm_pwd):
        self._logger = logger
        self._api = api
        self._sm_type = sm_type
        self._sm_data = sm_data
        sm_path = get_sm_path()
        self._sm_exe = str(sm_path/sm_type)
        self._sm_env = os.environ.copy()
        self._sm_env["LD_LIBRARY_PATH"] = str(sm_path)
        data_path = Path.home()/".tqsdk"/self._sm_type/sm_user
        data_path.mkdir(parents=True, exist_ok=True)
        self._sm_init = {
            # smi
            "DeviceId": hex(uuid.getnode()),
            "CertLog": str(data_path/"smi_cert.log"),
            "SdkLog": str(data_path/"smi_sdk.log"),
            # smf
            "Debug": True,
            "LogPath": str(data_path),
            "Timeout": 10000000000,
            # sms
            "LogFile":str(data_path/"syd.log"),
        }
        self._sm_cfg = {
            "UserId": sm_user,
            "Password": sm_pwd,
            "DataPath": str(data_path),
        }
        self._sm_addr = ""

    _log_level_map = {
        "trace": logging.DEBUG,
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warn": logging.WARNING,
        "error": logging.ERROR,
        "fatal": logging.FATAL,
        "panic": logging.CRITICAL,
    }

    def _process_log(self, line):
        # 仅处理完整的一行, 由于退出时先结束进程, 因此可能会输出半行日志
        if line and line[-1] == '\n' and line[:-1]:
            try:
                sm_log = json.loads(line[:-1])
            except json.JSONDecodeError:
                self._logger.error("sm error msg", error=line[:-1])
                return
            level = self._log_level_map[sm_log["level"]]
            self._logger.log(level, "sm log", pack=sm_log)
            if sm_log["msg"] == "NewConnection failed":
                err = sm_log["error"].split(":")[-1]
                if (self._sm_type in ["smi", "sms"] and err == "a000407") or (self._sm_type == "smf" and err == "-333700009"):
                    self._api._print(f"通知 {self._sm_cfg['UserId']}: 用户名或密码错误", level="ERROR")

    def _sync_err_logger(self, pipe):
        while True:
            line = pipe.readline()
            if not line:
                return
            # logging and _print is thread safe
            self._process_log(line.decode("utf-8"))

    async def _async_err_logger(self, reader):
        while not reader.at_eof():
            line = await reader.readline()
            self._process_log(line.decode("utf-8"))

    async def __aenter__(self):
        # Windows系统下不支持 connect_read_pipe 和 subprocess_exec
        if sys.platform.startswith("win"):
            self._sm_proc = subprocess.Popen([self._sm_exe, "-t", "dec"], stdin=PIPE, stdout=PIPE, stderr=PIPE, env=self._sm_env)
            self._logger_task = self._api._loop.run_in_executor(None, self._sync_err_logger, self._sm_proc.stderr)
            decrypt_coro = self._api._loop.run_in_executor(None, lambda: self._sm_proc.stdout.read())
            self._sm_proc.stdin.write(self._sm_data.encode("utf-8"))
            self._sm_proc.stdin.close()
            decrypt_out = await decrypt_coro
        else:
            rfd, wfd = os.pipe()
            # os.fdopen 返回的对象可以重复关闭, 而由于 fd 复用因此 os.pipe 返回的 fd 不能重复调用 os.close
            self._rf, self._wf = os.fdopen(rfd), os.fdopen(wfd)
            reader = asyncio.StreamReader()
            # 需要将 transport 保存下来, 在退出的时候关闭, 使得 _async_err_logger 退出
            self._rt, _ = await self._api._loop.connect_read_pipe(lambda: asyncio.StreamReaderProtocol(reader), self._rf)
            self._logger_task = self._api.create_task(self._async_err_logger(reader))
            self._sm_proc = await asyncio.create_subprocess_exec(self._sm_exe, "-t", "dec", stdin=PIPE, stdout=PIPE, stderr=self._wf, env=self._sm_env)
            decrypt_out, _ = await self._sm_proc.communicate(input=self._sm_data.encode("utf-8"))
        # ref:https://peps.python.org/pep-0448/
        # The keys in a dictionary remain in a right-to-left priority order, so {**{'a': 1}, 'a': 2, **{'a': 3}} evaluates to {'a': 3}.
        self._sm_cfg = {**json.loads(decrypt_out), **self._sm_cfg}
        return self

    async def get_addr(self):
        """无法启动时返回空字符串"""
        if sys.platform.startswith("win"):
            # subprocess.Popen 需要调用 poll 才会更新 returncode
            self._sm_proc.poll()
        # 初始由于执行了 -t dec 因此 returncode 不会为 None
        if self._sm_proc.returncode is not None:
            if sys.platform.startswith("win"):
                await self._logger_task
                self._sm_proc = subprocess.Popen([self._sm_exe, "-a", "localhost:0", "-t", self._sm_type], stdin=PIPE, stdout=PIPE, stderr=PIPE, env=self._sm_env)
                self._logger_task = self._api._loop.run_in_executor(None, self._sync_err_logger, self._sm_proc.stderr)
                addr_coro = self._api._loop.run_in_executor(None, lambda: self._sm_proc.stdout.readline())
                self._sm_proc.stdin.write(json.dumps(self._sm_init).encode("utf-8"))
                self._sm_proc.stdin.write(json.dumps(self._sm_cfg).encode("utf-8"))
                self._sm_proc.stdin.flush()
                self._sm_addr = (await addr_coro).decode("utf-8").strip()
            else:
                self._sm_proc = await asyncio.create_subprocess_exec(self._sm_exe, "-a", "localhost:0", "-t", self._sm_type, stdin=PIPE, stdout=PIPE, stderr=self._wf, env=self._sm_env)
                # 调用 write 但是不调用 drain, 避免死锁, 因为下面会 readline
                self._sm_proc.stdin.write(json.dumps(self._sm_init).encode("utf-8"))
                self._sm_proc.stdin.write(json.dumps(self._sm_cfg).encode("utf-8"))
                self._sm_addr = (await self._sm_proc.stdout.readline()).decode("utf-8").strip()
        return self._sm_addr

    async def __aexit__(self, exc_type, exc, tb):
        try:
            self._sm_proc.terminate()
        except ProcessLookupError:
            pass
        if sys.platform.startswith("win"):
            self._sm_proc.wait()
        else:
            await self._sm_proc.wait()
            # 会导致 _async_err_logger 遇到 eof
            self._rt.close()
            self._rf.close()
            self._wf.close()
        await self._logger_task


class NullContext(object):
    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        pass



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'mayanqiong'


from tqsdk.datetime import _get_expire_rest_days
from tqsdk.datetime_state import TqDatetimeState
from tqsdk.diff import _simple_merge_diff, _is_key_exist, _simple_merge_diff_and_collect_paths, _get_obj


class DataExtension(object):
    """
    为数据截面添加以下字段：
    {
        quotes: {
            *: {
                expire_rest_days: int
            }
        },
        trade: {
            *: {
                positions: {
                    *: {
                        'pos_long': int,
                        'pos_short': int,
                        'pos': int
                    }
                },
                orders: {
                    *: {
                        'is_dead': bool
                        'is_online': bool
                        'is_error': bool
                        'trade_price': float
                    }
                }
            }
        }
    }
    """

    def __init__(self, api):
        self._api = api
        self._data = {
            'trade': {},
            'quotes': {}  # 记录 quotes 的 expire_datetime，并不是每一次 diff 更新都会重新发送 expire_datetime，重新 expire_rest_days 时需要这个信息
        }  # 数据截面, 现在的功能只需要记录 trade
        self._diffs = []
        self._diffs_paths = set()
        self._prototype = {
            "trade": {
                "*": {
                    "orders": {
                        "*": {
                            "status": None,
                            "exchange_order_id": None
                        }
                    },
                    "trades": {
                        "*": None
                    },
                    "positions": {
                        "*": {
                            "pos_long_his": None,
                            "pos_long_today": None,
                            "pos_short_his": None,
                            "pos_short_today": None
                        }
                    }
                }
            }
        }

    async def _run(self, api_send_chan, api_recv_chan, md_send_chan, md_recv_chan):
        self._logger = self._api._logger.getChild("DataExtension")
        self._api_send_chan = api_send_chan
        self._api_recv_chan = api_recv_chan
        self._md_send_chan = md_send_chan
        self._md_recv_chan = md_recv_chan
        self._datetime_state = TqDatetimeState()
        md_task = self._api.create_task(self._md_handler())
        self._pending_peek = False  # True 表示收到下游的 peek_message ，并且没有发给过下游回复；False 表示发给过下游回复，没有 pending_peek_message
        self._pending_peek_md = False  # True 表示发给过上游 peek_message；False 表示对上游没有 pending_peek_message
        try:
            async for pack in api_send_chan:
                if "_md_recv" in pack:
                    if pack['aid'] == 'rtn_data':
                        self._pending_peek_md = False
                        await self._md_recv(pack)
                        await self._send_diff()
                    if self._pending_peek and self._pending_peek_md is False:
                        self._pending_peek_md = True
                        await self._md_send_chan.send({"aid": "peek_message"})
                elif pack["aid"] == "peek_message":
                    self._pending_peek = True
                    await self._send_diff()
                    if self._pending_peek and self._pending_peek_md is False:
                        self._pending_peek_md = True
                        await self._md_send_chan.send(pack)
                else:
                    await self._md_send_chan.send(pack)
        finally:
            await self._api._cancel_task(md_task)

    async def _md_handler(self):
        """0 接收上游数据包 """
        async for pack in self._md_recv_chan:
            pack["_md_recv"] = True
            await self._api_send_chan.send(pack)

    async def _md_recv(self, pack):
        """将行情数据和交易数据合并至 self._data """
        for d in pack.get("data", []):
            self._datetime_state.update_state(d)
            if d.get('quotes', None):
                _simple_merge_diff(
                    result=self._data['quotes'],
                    diff=d['quotes']
                )
            if d.get('trade', None):
                _simple_merge_diff_and_collect_paths(
                    result=self._data['trade'],
                    diff=d['trade'],
                    path=('trade', ),
                    diff_paths=self._diffs_paths,
                    prototype=self._prototype['trade']
                )
            self._diffs.append(d)

    def _generate_ext_diff(self):
        """"
        补充 quote, position 额外字段
        此函数在 send_diff() 才会调用， self._datetime_state.data_ready 一定为 True，
        调用 self._datetime_state.get_current_dt() 一定有正确的当前时间
        """
        for d in self._diffs:
            if d.get('quotes', None):
                self._update_quotes(d)
        pend_diff = {}
        _simple_merge_diff(pend_diff, self._get_positions_pend_diff())
        orders_set = set()  # 计算过委托单，is_dead、is_online、is_error
        orders_price_set = set()  # 根据成交计算哪些 order 需要重新计算平均成交价 trade_price
        for path in self._diffs_paths:
            if path[2] == 'orders':
                _, account_key, _, order_id, _ = path
                if (account_key, order_id) not in orders_set:
                    orders_set.add((account_key, order_id))
                    order = _get_obj(self._data, ['trade', account_key, 'orders', order_id])
                    if order:
                        pend_order = pend_diff.setdefault('trade', {}).setdefault(account_key, {}).setdefault('orders', {}).setdefault(order_id, {})
                        pend_order['is_dead'] = order['status'] == "FINISHED"
                        pend_order['is_online'] = order['exchange_order_id'] != "" and order['status'] == "ALIVE"
                        pend_order['is_error'] = order['exchange_order_id'] == "" and order['status'] == "FINISHED"
            elif path[2] == 'trades':
                _, account_key, _, trade_id = path
                trade = _get_obj(self._data, path)
                order_id = trade.get('order_id', '')
                if order_id:
                    orders_price_set.add(('trade', account_key, 'orders', order_id))
        for path in orders_price_set:
            _, account_key, _, order_id = path
            trade_price = self._get_trade_price(account_key, order_id)
            if trade_price == trade_price:
                pend_order = pend_diff.setdefault('trade', {}).setdefault(account_key, {}).setdefault('orders', {}).setdefault(order_id, {})
                pend_order['trade_price'] = trade_price
        self._diffs_paths = set()
        return pend_diff

    def _update_quotes(self, diff):
        for symbol in diff['quotes']:
            expire_datetime = self._data['quotes'].get(symbol, {}).get('expire_datetime', None)
            if expire_datetime and expire_datetime == expire_datetime:  # 排除 None 和 nan
                # expire_rest_days 距离到期日的剩余天数（自然日天数），正数表示距离到期日的剩余天数，0表示到期日当天，负数表示距离到期日已经过去的天数
                # 直接修改在 diff 里面的数据，当 diffs 里有多个对同个合约的修改时，保持数据截面的一致
                expire_rest_days = _get_expire_rest_days(expire_datetime, self._datetime_state.get_current_dt() / 1e9)
                diff['quotes'][symbol]['expire_rest_days'] = expire_rest_days

    def _get_positions_pend_diff(self):
        pend_diff = {}
        for account_key in self._data['trade']:
            positions = self._data['trade'][account_key].get('positions', {})
            for symbol, pos in positions.items():
                paths = [('trade', account_key, 'positions', symbol) + (key, )
                         for key in ['pos_long_his', 'pos_long_today', 'pos_short_his', 'pos_short_today']]
                if any([p in self._diffs_paths for p in paths]):
                    pos_long = pos['pos_long_his'] + pos['pos_long_today']
                    pos_short = pos['pos_short_his'] + pos['pos_short_today']
                    pend_diff.setdefault(account_key, {}).setdefault('positions', {})
                    pend_diff[account_key]['positions'][symbol] = {
                        'pos_long': pos_long,
                        'pos_short': pos_short,
                        'pos': pos_long - pos_short
                    }
        return {'trade': pend_diff} if pend_diff else {}

    def _get_trade_price(self, account_key, order_id):
        # 计算某个 order_id 对应的成交均价
        trades = self._data['trade'][account_key]['trades']
        trade_id_list = [t_id for t_id in trades.keys() if trades[t_id]['order_id'] == order_id]
        sum_volume = sum([trades[t_id]['volume'] for t_id in trade_id_list])
        if sum_volume == 0:
            return float('nan')
        else:
            sum_amount = sum([trades[t_id]['volume'] * trades[t_id]['price'] for t_id in trade_id_list])
            return sum_amount / sum_volume

    async def _send_diff(self):
        if self._datetime_state.data_ready and self._pending_peek and self._diffs:
            # 生成增量业务截面, 该截面包含补充的字段，只在真正需要给下游发送数据时，才将需要发送的数据放在 _diffs 中
            ext_diff = self._generate_ext_diff()
            rtn_data = {
                "aid": "rtn_data",
                "data": self._diffs + [ext_diff],
            }
            self._diffs = []
            self._pending_peek = False
            await self._api_recv_chan.send(rtn_data)



#!usr/bin/env python3
#-*- coding:utf-8 -*-

from tqsdk import TqApi, TqAuth, tafunc
from tqsdk.ta import *

api = TqApi(auth=TqAuth("快期账户", "账户密码"))

underlying_quote = api.get_quote("SHFE.cu2009")
klines = api.get_kline_serial('SHFE.cu2009', 24 * 60 * 60, 20)
v = tafunc.get_his_volatility(klines, underlying_quote)
print("历史波动率:", v)

quote = api.get_quote("SHFE.cu2009C44000")
bs_serise = BS_VALUE(klines, quote, 0.025)
print("理论价:", list(round(bs_serise['bs_price'], 2)))


klines2 = api.get_kline_serial(["SHFE.cu2009C44000", "SHFE.cu2009"], 24 * 60 * 60, 20)

values = OPTION_VALUE(klines2, quote)
print("内在价值:", list(values["intrins"]))
print("时间价值:", list(values["time"]))

impv = OPTION_IMPV(klines2, quote, 0.025)
print("隐含波动率:", list(round(impv['impv'] * 100, 2)))

greeks = OPTION_GREEKS(klines2, quote, 0.025, impv['impv'])
print("delta:", list(greeks["delta"]))
print("theta:", list(greeks["theta"]))
print("gamma:", list(greeks["gamma"]))
print("vega:", list(greeks["vega"]))
print("rho:", list(greeks["rho"]))

api.close()



#!usr/bin/env python3
# -*- coding:utf-8 -*-

__author__ = 'mayanqiong'


from abc import abstractmethod


class TqRiskRule(object):

    def __init__(self, api, account=None):
        # 记录必要参数， 每个风控规则都需要继承这个基类
        self._api = api
        account = self._api._account._check_valid(account)
        if account is None:
            raise Exception(f"多账户模式下, 需要指定账户实例 account")
        self._account = account
        self._account_key = self._account._account_key

    @abstractmethod
    def _could_insert_order(self, pack) -> (bool, str):  # 是否可以下单
        pass

    @abstractmethod
    def _on_insert_order(self, pack):
        pass

    @abstractmethod
    def _on_settle(self):
        pass


class TqRuleOpenCountsLimit(TqRiskRule):
    """
    风控规则类 - 交易日内开仓次数限制。

    此功能为 TqSdk 专业版提供，如需使用此功能，可以点击 `天勤量化专业版 <https://www.shinnytech.com/tqsdk_professional/>`_ 申请试用或购买
    """

    def __init__(self, api, open_counts_limit, symbol, account=None):
        """
        Args:
            api (TqApi): TqApi 实例

            open_volumes_limit (int): 交易日内开仓手数上限

            symbol (str/list of str): 负责限制的合约代码或合约代码列表.
                * str: 一个合约代码
                * list of str: 合约代码列表

            account  (TqAccount/TqKq/TqSim): [可选] 指定发送下单指令的账户实例, 多账户模式下，该参数必须指定

        Example1::

            from tqsdk import TqApi
            from tqsdk.risk_rule import TqRuleOpenCountsLimit

            api = TqApi(auth=TqAuth("快期账户", "账户密码"))

            rule = TqRuleOpenCountsLimit(api, open_counts_limit=10, symbol="DCE.m2112")  # 创建风控规则实例
            api.add_risk_rule(rule)  # 添加风控规则

            quote = api.get_quote("DCE.m2112")
            try:
                # 每次最新价变动，下一笔订单，直到超过开仓次数风控限制
                while True:
                    api.wait_update()
                    if api.is_changing(quote, ['last_price']):
                        order = api.insert_order(symbol="DCE.m2112", direction="BUY", offset="OPEN", volume=1)
                        while order.status != "FINISHED":
                            api.wait_update()
            except TqRiskRuleError as e:
                print('!!!', e)
            api.close()

        """
        super(TqRuleOpenCountsLimit, self).__init__(api=api, account=account)
        if open_counts_limit < 0:
            raise Exception("参数 open_volumes_limit 必须大于 0 的数字")
        self.open_counts_limit = open_counts_limit
        self.symbol_list = [symbol] if isinstance(symbol, str) else symbol
        self.data = {s: 0 for s in self.symbol_list}
        for order_id, order in self._api._data.get('trade', {}).get(self._account_key, {}).get('orders', {}).items():
            symbol = order["exchange_id"] + "." + order["instrument_id"]
            if order["offset"] == "OPEN" and symbol in self.data:
                self.data[symbol] += 1

    def _could_insert_order(self, pack) -> {bool, str}:
        if pack['account_key'] == self._account_key:
            symbol = pack["exchange_id"] + "." + pack["instrument_id"]
            if pack["offset"] == "OPEN" and symbol in self.symbol_list:
                if self.data[symbol] + 1 > self.open_counts_limit:
                    return False, f"触发风控规则，合约 {symbol} 开仓到达交易日内开仓次数限制 {self.open_counts_limit}, " \
                                  f"已下单次数 {self.data[symbol]}"
        return True, ""

    def _on_insert_order(self, pack):
        if pack['account_key'] == self._account_key:
            symbol = pack["exchange_id"] + "." + pack["instrument_id"]
            if pack["offset"] == "OPEN" and symbol in self.symbol_list:
                self.data[symbol] += 1

    def _on_settle(self):
        for k in self.data.keys():
            self.data[k] = 0


class TqRuleOpenVolumesLimit(TqRiskRule):
    """
    风控规则类 - 交易日内开仓手数限制

    此功能为 TqSdk 专业版提供，如需使用此功能，可以点击 `天勤量化专业版 <https://www.shinnytech.com/tqsdk_professional/>`_ 申请试用或购买
    """

    def __init__(self, api, open_volumes_limit, symbol, account=None):
        """
        Args:
            api (TqApi): TqApi 实例

            open_volumes_limit (int): 交易日内开仓手数上限

            symbol (str/list of str): 负责限制的合约代码或合约代码列表.
                * str: 一个合约代码
                * list of str: 合约代码列表

            account  (TqAccount/TqKq/TqSim): [可选] 指定发送下单指令的账户实例, 多账户模式下，该参数必须指定

        Example1::

            from tqsdk import TqApi
            from tqsdk.risk_rule import TqRuleOpenVolumesLimit

            api = TqApi(auth=TqAuth("快期账户", "账户密码"))

            rule = TqRuleOpenVolumesLimit(api, open_volumes_limit=10, symbol="DCE.m2112")  # 创建风控规则实例
            api.add_risk_rule(rule)  # 添加风控规则

            # 下单 5 手，不会触发风控规则
            order1 = api.insert_order(symbol="DCE.m2112", direction="BUY", offset="OPEN", volume=5)
            while order1.status != "FINISHED":
                api.wait_update()

            # 继续下单 8 手，会触发风控规则
            order2 = api.insert_order(symbol="DCE.m2112", direction="BUY", offset="OPEN", volume=8)
            while order2.status != "FINISHED":
                api.wait_update()
            api.close()


        Example2::


            from tqsdk import TqApi, TqKq, TqRiskRuleError
            from tqsdk.risk_rule import TqRuleOpenVolumesLimit

            account = TqKq()
            api = TqApi(account=account, auth=TqAuth("快期账户", "账户密码"))

            rule = TqRuleOpenVolumesLimit(api, open_volumes_limit=10, symbol="DCE.m2112", account=account)  # 创建风控规则实例
            api.add_risk_rule(rule)  # 添加风控规则

            try:
                # 下单 11 手，触发风控规则
                order1 = api.insert_order(symbol="DCE.m2112", direction="BUY", offset="OPEN", volume=11)
                while order1.status != "FINISHED":
                    api.wait_update()
            except TqRiskRuleError as e:
                print("!!!", e)

            api.close()
        """
        super(TqRuleOpenVolumesLimit, self).__init__(api=api, account=account)
        if open_volumes_limit < 0:
            raise Exception("参数 open_volumes_limit 必须大于 0 的数字")
        self.open_volumes_limit = open_volumes_limit
        self.symbol_list = [symbol] if isinstance(symbol, str) else symbol
        self.data = {s: 0 for s in self.symbol_list}
        for trade_id, trade in self._api._data.get('trade', {}).get(self._account_key, {}).get('trades', {}).items():
            symbol = trade["exchange_id"] + "." + trade["instrument_id"]
            if trade["offset"] == "OPEN" and symbol in self.data:
                self.data[symbol] += trade["volume"]

    def _could_insert_order(self, pack) -> {bool, str}:
        if pack['account_key'] == self._account_key:
            symbol = pack["exchange_id"] + "." + pack["instrument_id"]
            if pack["offset"] == "OPEN" and symbol in self.symbol_list:
                if self.data[symbol] + pack["volume"] > self.open_volumes_limit:
                    return False, f"触发风控规则，合约 {symbol} 开仓到达交易日内开仓手数限制 {self.open_volumes_limit}, " \
                                  f"已下单手数 {self.data[symbol]}, 即将下单手数 {pack['volume']}"
        return True, ""

    def _on_insert_order(self, pack):
        if pack['account_key'] == self._account_key:
            symbol = pack["exchange_id"] + "." + pack["instrument_id"]
            if pack["offset"] == "OPEN" and symbol in self.symbol_list:
                self.data[symbol] += pack["volume"]

    def _on_settle(self):
        for k in self.data.keys():
            self.data[k] = 0


class TqRuleAccOpenVolumesLimit(TqRiskRule):
    """
    风控规则类 - 累计开仓手数限制。

    限制合约开仓手数之和。

    此功能为 TqSdk 专业版提供，如需使用此功能，可以点击 `天勤量化专业版 <https://www.shinnytech.com/tqsdk_professional/>`_ 申请试用或购买

    """

    def __init__(self, api, open_volumes_limit, symbol, account=None):
        """
        Args:
            api (TqApi): TqApi 实例


            open_volumes_limit (int): 交易日内开仓手数之和上限

            symbol (str/list of str): 负责限制的合约代码或合约代码列表.
                * str: 一个合约代码
                * list of str: 合约代码列表

            account  (TqAccount/TqKq/TqSim): [可选] 指定发送下单指令的账户实例, 多账户模式下，该参数必须指定

        Example::

            from tqsdk import TqApi, TqKq, TqRiskRuleError
            from tqsdk.risk_rule import TqRuleAccOpenVolumesLimit

            account = TqKq()
            api = TqApi(account=account, auth=TqAuth("快期账户", "账户密码"))

            quote = api.get_quote("SSE.000300")
            call_in, call_at, call_out = api.query_all_level_finance_options("SSE.000300", quote.last_price, "CALL", nearbys=0)
            put_in, put_at, put_out = api.query_all_level_finance_options("SSE.000300", quote.last_price, "PUT", nearbys=0)
            near_symbols = call_in + call_at + call_out + put_in + put_at + put_out  # 找到所有当月期权合约

            symbols = api.query_options("SSE.000300", expired=False)  # 找到所有中金所期权合约

            # 规则1: 中金所当月期权合约日内开仓不超过 100 手
            # 规则2: 中金所所有期权合约日内合约开仓不超过 200 手
            rule1 = TqRuleAccOpenVolumesLimit(api, open_volumes_limit=100, symbol=near_symbols, account=account)  # 创建风控规则实例
            rule2 = TqRuleAccOpenVolumesLimit(api, open_volumes_limit=200, symbol=symbols, account=account)  # 创建风控规则实例
            api.add_risk_rule(rule1)  # 添加风控规则
            api.add_risk_rule(rule2)  # 添加风控规则

            try:
                # 下单 101 手，触发风控规则
                order1 = api.insert_order(symbol="CFFEX.IO2111-C-4900", direction="BUY", offset="OPEN", volume=101, limit_price=35.6)
                while order1.status != "FINISHED":
                    api.wait_update()
            except TqRiskRuleError as e:
                print("!!!", e)  # 报错，当月期权合约日内合约开仓不超过 100 手, 已下单次数 0

            api.close()
        """
        super(TqRuleAccOpenVolumesLimit, self).__init__(api=api, account=account)
        if open_volumes_limit < 0:
            raise Exception("参数 open_volumes_limit 必须大于 0 的数字")
        self.open_volumes_limit = open_volumes_limit
        self.open_volumes = 0  # 所有合约日内合约开仓手数
        self.symbol_list = [symbol] if isinstance(symbol, str) else symbol
        for trade_id, trade in self._api._data.get('trade', {}).get(self._account_key, {}).get('trades', {}).items():
            if trade["offset"] == "OPEN" and f'{trade["exchange_id"]}.{trade["instrument_id"]}' in self.symbol_list:
                self.open_volumes += trade['volume']

    def _could_insert_order(self, pack) -> {bool, str}:
        symbol = pack["exchange_id"] + "." + pack["instrument_id"]
        if pack['account_key'] == self._account_key and symbol in self.symbol_list:  # 当前账户下单
            if pack["offset"] == "OPEN" and self.open_volumes + pack['volume'] > self.open_volumes_limit:
                return False, f"触发风控规则，所有合约日内合约开仓不超过 {self.open_volumes_limit} 手, 已下单手数 {self.open_volumes}"
        return True, ""

    def _on_insert_order(self, pack):
        symbol = pack["exchange_id"] + "." + pack["instrument_id"]
        if pack['account_key'] == self._account_key and symbol in self.symbol_list:  # 当前账户下单
            if pack["offset"] == "OPEN":  # 开仓单把手数累加
                self.open_volumes += pack['volume']

    def _on_settle(self):
        self.open_volumes = 0



#!usr/bin/env python3
#-*- coding:utf-8 -*-
__author__ = 'yanqiong'

import copy
import weakref
from collections.abc import MutableMapping


class Entity(MutableMapping):
    def _instance_entity(self, path):
        self._path = path
        self._listener = weakref.WeakSet()

    def __setitem__(self, key, value):
        return self.__dict__.__setitem__(key, value)

    def __delitem__(self, key):
        return self.__dict__.__delitem__(key)

    def __getitem__(self, key):
        return self.__dict__.__getitem__(key)

    def __iter__(self):
        return iter({k: v for k, v in self.__dict__.items() if not k.startswith("_")})

    def __len__(self):
        return len({k: v for k, v in self.__dict__.items() if not k.startswith("_")})

    def __str__(self):
        return str({k: v for k, v in self.__dict__.items() if not k.startswith("_")})

    def __repr__(self):
        return '{}, D({})'.format(super(Entity, self).__repr__(),
                                  {k: v for k, v in self.__dict__.items() if not k.startswith("_")})

    def copy(self):
        return copy.copy(self)



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'limin'

import pandas as pd
import datetime
from contextlib import closing
from tqsdk import TqApi, TqAuth, TqBacktest, BacktestFinished, TargetPosTask
from tqsdk.tafunc import sma, ema2, trma
from sklearn.ensemble import RandomForestClassifier

pd.set_option('display.max_rows', None)  # 设置Pandas显示的行数
pd.set_option('display.width', None)  # 设置Pandas显示的宽度

'''
应用随机森林对某交易日涨跌情况的预测(使用sklearn包)
参考:https://www.joinquant.com/post/1571
注: 该示例策略仅用于功能示范, 实盘时请根据自己的策略/经验进行修改
'''

symbol = "SHFE.ru1811"  # 交易合约代码
close_hour, close_minute = 14, 50  # 预定收盘时间(因为真实收盘后无法进行交易, 所以提前设定收盘时间)


def get_prediction_data(klines, n):
    """获取用于随机森林的n个输入数据(n为数据长度): n天中每天的特征参数及其涨跌情况"""
    close_prices = klines.close[- 30 - n:]  # 获取本交易日及以前的收盘价(此时在预定的收盘时间: 认为本交易日已收盘)
    # 计算所需指标
    sma_data = sma(close_prices, 30, 0.02)[-n:]  # SMA指标, 函数默认时间周期参数:30
    wma_data = ema2(close_prices, 30)[-n:]  # WMA指标
    mom_data = trma(close_prices, 30)[-n:]  # MOM指标
    x_all = list(zip(sma_data, wma_data, mom_data))  # 样本特征组
    y_all = list(klines.close.iloc[i] >= klines.close.iloc[i - 1] for i in list(reversed(range(-1, -n - 1, -1))))  # 样本标签组
    # x_all:            大前天指标 前天指标 昨天指标 (今天指标)
    # y_all:   (大前天)    前天     昨天    今天      -明天-
    # 准备算法需要用到的数据
    x_train = x_all[: -1]  # 训练数据: 特征
    x_predict = x_all[-1]  # 预测数据(用本交易日的指标预测下一交易日的涨跌)
    y_train = y_all[1:]  # 训练数据: 标签 (去掉第一个数据后让其与指标隔一位对齐(例如: 昨天的特征 -> 对应预测今天的涨跌标签))

    return x_train, y_train, x_predict


predictions = []  # 用于记录每次的预测结果(在每个交易日收盘时用收盘数据预测下一交易日的涨跌,并记录在此列表里)
api = TqApi(backtest=TqBacktest(start_dt=datetime.date(2018, 7, 2), end_dt=datetime.date(2018, 9, 26)), auth=TqAuth("快期账户", "账户密码"))
quote = api.get_quote(symbol)
klines = api.get_kline_serial(symbol, duration_seconds=24 * 60 * 60)  # 日线
target_pos = TargetPosTask(api, symbol)
with closing(api):
    try:
        while True:
            while not api.is_changing(klines.iloc[-1], "datetime"):  # 等到达下一个交易日
                api.wait_update()
            while True:
                api.wait_update()
                # 在收盘后预测下一交易日的涨跌情况
                if api.is_changing(quote, "datetime"):
                    now = datetime.datetime.strptime(quote.datetime, "%Y-%m-%d %H:%M:%S.%f")  # 当前quote的时间
                    # 判断是否到达预定收盘时间: 如果到达 则认为本交易日收盘, 此时预测下一交易日的涨跌情况, 并调整为对应仓位
                    if now.hour == close_hour and now.minute >= close_minute:
                        # 1- 获取数据
                        x_train, y_train, x_predict = get_prediction_data(klines, 75)  # 参数1: K线, 参数2:需要的数据长度

                        # 2- 利用机器学习算法预测下一个交易日的涨跌情况
                        # n_estimators 参数: 选择森林里（决策）树的数目; bootstrap 参数: 选择建立决策树时，是否使用有放回抽样
                        clf = RandomForestClassifier(n_estimators=30, bootstrap=True)
                        clf.fit(x_train, y_train)  # 传入训练数据, 进行参数训练
                        predictions.append(bool(clf.predict([x_predict])))  # 传入测试数据进行预测, 得到预测的结果

                        # 3- 进行交易
                        if predictions[-1] == True:  # 如果预测结果为涨: 买入
                            print(quote.datetime, "预测下一交易日为 涨")
                            target_pos.set_target_volume(10)
                        else:  # 如果预测结果为跌: 卖出
                            print(quote.datetime, "预测下一交易日为 跌")
                            target_pos.set_target_volume(-10)
                        break

    except BacktestFinished:  # 回测结束, 获取预测结果，统计正确率
        klines["pre_close"] = klines["close"].shift(1)  # 增加 pre_close(上一交易日的收盘价) 字段
        klines = klines[-len(predictions) + 1:]  # 取出在回测日期内的K线数据
        klines["prediction"] = predictions[:-1]  # 增加预测的本交易日涨跌情况字段(向后移一个数据目的: 将 本交易日对应下一交易日的涨跌 调整为 本交易日对应本交易日的涨跌)
        results = (klines["close"] - klines["pre_close"] >= 0) == klines["prediction"]

        print(klines)
        print("----回测结束----")
        print("预测结果正误:\n", results)
        print("预测结果数目统计: 总计", len(results),"个预测结果")
        print(pd.value_counts(results))
        print("预测的准确率:")
        print((pd.value_counts(results)[True]) / len(results))



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'mayanqiong'

from tqsdk.datetime import _timestamp_nano_to_datetime
from tqsdk.objs import Quote


def _symbols_to_quotes(symbols, keys=set(Quote(None).keys())):
    """将 symbols 转为 quotes，只输出 keys 包括的字段"""
    result = symbols.get("result", {})
    quotes = {}
    for k in result:
        for symbol in result[k]:
            quote = quotes.setdefault(symbol["instrument_id"], {})
            quote.update(_convert_symbol_to_quote(symbol, keys))
            if symbol.get("underlying"):
                for edge in symbol["underlying"]["edges"]:
                    underlying_symbol = edge["node"]
                    if "underlying_symbol" in keys:
                        quote["underlying_symbol"] = underlying_symbol["instrument_id"]
                    underlying_quote = quotes.setdefault(underlying_symbol["instrument_id"], {})
                    underlying_quote.update(_convert_symbol_to_quote(underlying_symbol, keys))
                    # 为期权合约补充 delivery_year delivery_month 商品期权根据标的赋值；金融期权与 exercise_year exercise_month 相同
                    # 为期权补充 delivery_year delivery_month 完全是为了兼容旧版合约服务
                    for key in ["delivery_year", "delivery_month"]:
                        if key in keys and symbol["class"] == "OPTION":
                            if symbol["exchange_id"] in ["DCE", "CZCE", "SHFE", "GFEX"]:
                                quote[key] = underlying_quote[key]
                            if symbol["exchange_id"] == "CFFEX" and "last_exercise_datetime" in symbol:
                                if key == "delivery_year":
                                    quote[key] = _timestamp_nano_to_datetime(symbol["last_exercise_datetime"]).year
                                else:
                                    quote[key] = _timestamp_nano_to_datetime(symbol["last_exercise_datetime"]).month
    for k in quotes:
        if quotes[k].get("ins_class", "") == "COMBINE":
            # 为组合合约补充 volume_multiple
            leg1_quote = quotes.get(quotes[k].get("leg1_symbol", ""), {})
            if leg1_quote:
                if leg1_quote.get("volume_multiple"):
                    quotes[k]["volume_multiple"] = leg1_quote["volume_multiple"]
    return quotes


def _convert_symbol_to_quote(symbol, keys):
    quote = {}
    for key in keys:
        if key == "leg1_symbol" and "leg1" in symbol:
            quote["leg1_symbol"] = symbol["leg1"]["instrument_id"]
        elif key == "leg2_symbol" and "leg2" in symbol:
            quote["leg2_symbol"] = symbol["leg2"]["instrument_id"]
        elif key == "ins_class" and "class" in symbol:
            quote["ins_class"] = symbol["class"]
        elif key == "option_class" and "call_or_put" in symbol:
            quote["option_class"] = symbol["call_or_put"]
        elif key == "volume_multiple" and "index_multiple" in symbol:
            quote["volume_multiple"] = symbol["index_multiple"]
        elif key == "expire_datetime" and symbol.get("expire_datetime"):
            quote["expire_datetime"] = symbol["expire_datetime"] / 1e9
        elif key == "last_exercise_datetime" and symbol.get("last_exercise_datetime"):
            quote["last_exercise_datetime"] = symbol["last_exercise_datetime"] / 1e9
        elif key == "exercise_year" and symbol.get("last_exercise_datetime"):
            quote["exercise_year"] = _timestamp_nano_to_datetime(symbol["last_exercise_datetime"]).year
        elif key == "exercise_month" and symbol.get("last_exercise_datetime"):
            quote["exercise_month"] = _timestamp_nano_to_datetime(symbol["last_exercise_datetime"]).month
        elif key == "pre_settlement" and "settlement_price" in symbol:
            quote["pre_settlement"] = symbol["settlement_price"]
        elif key in symbol:
            quote[key] = symbol[key]
    return quote